#include <exception>
#include <iostream>
#include <mimas/image_mesaoutput.h>
#include <mimas/image_v4linput.h>
#include <popt.h>

using namespace mimas;
using namespace std;

/** Compute average of image.
    @param image Image to compute average for.
    @return average value. */
template< typename T >
int average( const mimas::image< T > &img )
{
  int n = img.rawData().num_elements();
  const T *p = img.rawData().data();
  int sum = 0;
  for ( int i=0; i<n; i++ )
    sum += *p++;
  return sum / n;
}

/// Class for adjusting brightness of camera.
class adjust_camera
{
 public:
  /** Constructor.
      @param _nominalAverage Desired average value of images.
      @param _cam Access to camera
      @param _contrast Contrast to be set */
  adjust_camera( unsigned char _nominalAverage,
                 image_v4linput< rgba< unsigned char > > &_cam,
                 int _contrast );
  /** Process one image.
      The image is used for changing the sensitivity of the camera.
      @image Image to be processed. */
  void process( const image< unsigned char > &img );
  /// Get current brightness-parameter for camera.
  int getBrightness(void) const { return brightness; }
  /// Get current step-size for adjusting brightness-parameter.
  int getBrightnessStep(void) const { return brightnessStep; }
 protected:
  /** Adjust value of camera.
      @param up Indicate, wether to increase or decrease value.
      @param value Pointer to value.
      @param step Pointer to corresponding step-size. */
  void adjust( bool up, int adjust_camera::* value,
               int adjust_camera::* step );
  /// Pointer to camera-object.
  image_v4linput< rgba< unsigned char > > *cam;
  /// Nominal average.
  unsigned char nominalAverage;
  /// Step-size for adjusting brightness.
  int brightnessStep;
  /// Brightness-parameter for camera.
  int brightness;
  /// Contrast-parameter for camera.
  int contrast;
};

adjust_camera::adjust_camera( unsigned char _nominalAverage,
                              image_v4linput< rgba< unsigned char > >
                              &_cam,
                              int _contrast ):
  cam(&_cam),
  nominalAverage(_nominalAverage),
  brightnessStep(8192), brightness(32768), contrast(_contrast)
{
  cam->setSensivity( brightness, 32768, 32768, contrast );
}

void adjust_camera::process( const image< unsigned char > &img )
{
  int avg = average( img );
  adjust( avg < nominalAverage,
          &adjust_camera::brightness, &adjust_camera::brightnessStep );
  cam->setSensivity( brightness, 32768, 32768, contrast );
}

void adjust_camera::adjust( bool up, int adjust_camera::* value,
                            int adjust_camera::* step )
{
  if ( this->*step != 0 ) {
    if ( this->*step > 0 ) {
      if ( up )
        this->*value += this->*step;
    else {
      this->*step = -( this->*step / 2 );
      this->*value += this->*step;
    };
    } else if ( this->*step < 0 ) {
      if ( !up )
        this->*value += this->*step;
      else {
        this->*step = -( this->*step / 2 );
        this->*value += this->*step;
      };
    };
    if ( this->*value < 0 ) {
      this->*value = 0;
      this->*step = 0;
    };
    if ( this->*value > 65535 ) {
      this->*value = 65535;
      this->*step = 0;
    };
  };
}

int main( int argc, char *argv[] )
{
  int
    retVal = 0,
    channel = 0,
    average = -1,
    contrast = 32768,
    width = -1,
    height = -1;

  const char *device = "/dev/video";

  struct poptOption optionsTable[] = {
    { "device", 'd', POPT_ARG_STRING | POPT_ARGFLAG_SHOW_DEFAULT,
      &device, 0, "Specify video4linux-device", "device-name" },
    { "channel", 'n', POPT_ARG_INT | POPT_ARGFLAG_SHOW_DEFAULT,
      &channel, 0, "Specify video-channel", "channel-number" },
    { "width", 'w', POPT_ARG_INT,
      &width, 0, "Force width of camera-image", "width in pixel" },
    { "height", 'h', POPT_ARG_INT,
      &height, 0, "Force height of camera-image", "height in pixel" },
    { "contrast", 'c', POPT_ARG_INT | POPT_ARGFLAG_SHOW_DEFAULT,
      &contrast, 0, "Contrast for video-input (not supported by some devices) "
      "[0..65535]", "contrast" },
    { "average", 'a', POPT_ARG_INT, &average, 0,
      "Nominal value for average gray-level [0..255]", "nominal value" },
    POPT_AUTOHELP
    { NULL, 0, 0, NULL, 0 } };
  
  poptContext optCon( poptGetContext( NULL, argc, (const char **)argv,
                                      optionsTable, 0 ) );

  try {

    // Parse command-line.
    char c;
    while ( ( c = poptGetNextOpt( optCon ) ) >= 0 );

    // Throw exception on error.
    MMERROR( c >= -1, mimasexception, ,
             poptBadOption( optCon, POPT_BADOPTION_NOALIAS ) << ": "
             << poptStrerror( c ) );
    MMERROR( average != -1, mimasexception, ,
             "No nominal value for average was specified." );
    MMERROR( average >= 0 && average <= 255, mimasexception, ,
             "Nominal average must be in [0..255], but was " << average
             << "." );
    MMERROR( contrast >= 0 && contrast <= 65535, mimasexception, ,
             "Contrast must be in [0..65535], but was " << contrast
             << "." );

    x11_display display;
    image_v4linput< rgba< unsigned char > > cam( device, channel,
                                                       width, height );
    cam.setSensivity( 32768, 32768, 32768, 32768 );
    image_mesaoutput< rgba< unsigned char > > win( &display,
                                                   cam.getWidth(),
                                                   cam.getHeight() );
    adjust_camera adjust( average, cam, contrast );
    while ( adjust.getBrightnessStep() != 0 ) {
      image< rgba< unsigned char > > img;
      cam >> img;
      win << img;
      adjust.process( image< unsigned char >( img ) );
      cout << "Brightness was set to " << adjust.getBrightness() << "." << endl;
    };

    cout << endl << "Final result is " << adjust.getBrightness() << "."
         << endl;
    
  } catch ( exception &e ) {
    cerr << e.what() << endl;
    // Display command-line information.
    poptPrintUsage( optCon, stderr, 0 );
    retVal = 1;
  };

  // Call destructor of command-line parser.
  poptFreeContext( optCon );

  return retVal;
}
