#ifndef __CALIBRATION_H
#define __CALIBRATION_H

#include <boost/shared_ptr.hpp>
#include <mimas/object.h>
#include <mimas/xml_reference.h>
#include "mi_pose.hh"

/** Calibration.
    Object holding calibration data.
    @author Jan Wedekind (j.wedekind@shu.ac.uk) */
class calibration: public mimas::object
{
public:
  /** Constructor.
      Read calibration data from XML-document.
      @param calibration_node Node with calibration-data. */
  calibration( const mimas::xml_reference< XalanElement >
               &calibration_node ) throw (mimas::mimasexception);
  /// Get width of camera-images.
  int get_width(void) const { return width; }
  /// Get height of camera-images.
  int get_height(void) const { return height; }
  /// Sets the width of the camera images.
  void set_width( int _width ) { width = _width; }
  /// Sets the height of the camera images.
  void set_height( int _height ) { height = _height; }
  /// X-coordinate of middle of image.
  int get_middle_x(void) const { return width / 2; }
  /// Y-coordinate of middle of image.
  int get_middle_y(void) const { return height / 2; }
  /** Get width or height of projected pixel.
      @param steps Offset in steps. */
  double get_pixel_size(void) const
    { return pixel_size / get_scale(); }
  /// Distance between optical center and focussed plane.
  double get_focus_distance(void) const
    { return focus_distance + focus_shift * steps; }
  /// Update position of camera drive.
  void setSteps( int _steps ) { steps = _steps; }
  /// Get last stored position of camera drive.
  int getSteps(void) const { return steps; }
  /// Scaling introduced by offset of focus drive.
  double get_scale(void) const
    { return 1.0 + scale_change * steps; }
protected:
  /// Current position of focus-drive.
  int steps;
  /// Width of camera-images.
  int width;
  /// Height of camera-images.
  int height;
  /// Width and height of projected pixel.
  double pixel_size;
  /// Default-distance between focussed plane and optical center of camera.
  double focus_distance;
  /// Focus-shift per step of focus-drive's motor.
  double focus_shift;
  /// Scale change per step of focus-drive's motor.
  double scale_change;
};

///
typedef boost::shared_ptr< calibration > calibration_ptr;

#endif
