// #define DESPECKLE

#include <exception>
#include <iomanip>
#include <iostream>
#ifdef DESPECKLE
#include <mimas/image_funcs.h>
#endif
#include <mimas/image_mesaoutput.h>
#include <mimas/image_filesinput.h>
#include <mimas/image_fileoutput.h>
#include <mimas/image_v4linput.h>
#include <popt.h>
#include <sstream>
#include "tools.hh"

using namespace boost;
using namespace mimas;
using namespace std;

int numFeatures( const image< bool > &binary )
{
  int retVal = 0;
  const bool *p = binary.rawData().data();
  for ( int y=0; y<(signed)binary.getHeight(); y++ )
    for ( int x=0; x<(signed)binary.getWidth(); x++ )
      if ( *p++ ) retVal++;
  return retVal;
}

int main( int argc, char *argv[] )
{
  int
    retVal = 0,
    selectedInput = 0,
    savePics = 0,
    gstreamerSHMKey = 333,
    gstreamerSEMKey = 333,
    channel = 0,
    brightness = 32768,
    contrast = 32768,
    threshold = -1,
    count = 0,
    fringe = 0,
    sample = 1,
    width = -1,
    height = -1,
    shutter = -1,
    gain = -1,
    balance = -1;

  const char *device = NULL;

  struct poptOption optionsTable[] = {
    { "gstreamer", 'g', POPT_ARG_VAL, &selectedInput, 1,
      "Select gstreamer input for reading images", 0 },
    { "shm", '\0', POPT_ARG_INT | POPT_ARGFLAG_SHOW_DEFAULT, &gstreamerSHMKey,
      0, "shared memory key for gstreamer input", "integer" },
    { "sem", '\0', POPT_ARG_INT | POPT_ARGFLAG_SHOW_DEFAULT, &gstreamerSEMKey,
      0, "semaphore key for gstreamer input", "integer" },
    { "v4l", 'v', POPT_ARG_VAL, &selectedInput, 2,
      "Select video4linux for reading images", 0 },
    { "brightness", '\0', POPT_ARG_INT | POPT_ARGFLAG_SHOW_DEFAULT,
      &brightness, 0, "Brightness for video4linux-input [0..65535]",
      "brightness" },
    { "contrast", '\0', POPT_ARG_INT | POPT_ARGFLAG_SHOW_DEFAULT,
      &contrast, 0, "Contrast for video4linux-input [0..65535]", "contrast" },
    { "video1394", '\0', POPT_ARG_VAL, &selectedInput, 3,
      "Select firewire digital camera for reading images", 0 },
    { "shutter", '\0', POPT_ARG_INT | POPT_ARGFLAG_SHOW_DEFAULT,
      &shutter, 0, "Shutter for video1394-input", "shutter" },
    { "gain", '\0', POPT_ARG_INT | POPT_ARGFLAG_SHOW_DEFAULT,
      &gain, 0, "Gain for video1394-input", "gain" },
    { "balance", '\0', POPT_ARG_INT | POPT_ARGFLAG_SHOW_DEFAULT,
      &balance, 0, "White balance for video1394-input", "balance" },
    { "device", '\0', POPT_ARG_STRING,
      &device, 0, "Specify video4linux/video1394-device", "device-name" },
    { "channel", '\0', POPT_ARG_INT | POPT_ARGFLAG_SHOW_DEFAULT,
      &channel, 0, "Specify video4linux-channel/video1394-node",
      "channel-number" },
    { "width", 'w', POPT_ARG_INT,
      &width, 0, "Force width of camera-image", "width in pixel" },
    { "height", 'h', POPT_ARG_INT,
      &height, 0, "Force height of camera-image", "height in pixel" },
    { "threshold", 't', POPT_ARG_INT, &threshold, 0,
      "Edge-detector's threshold for the gradient-values", "threshold" },
    { "fringe", 'f', POPT_ARG_INT | POPT_ARGFLAG_SHOW_DEFAULT, &fringe, 0,
      "Cut border-pixels from the image to remove features caused by them",
      "number of pixels" },
    { "sample", 's', POPT_ARG_INT | POPT_ARGFLAG_SHOW_DEFAULT, &sample, 0,
      "Sub-sample image before processing it", "sampling rate" },
    { "savepics", '\0', POPT_ARG_VAL, &savePics, 1,
      "Save pictures for debugging-purposes (edge*.ppm)", 0 },
    POPT_AUTOHELP
    { NULL, 0, 0, NULL, 0 } };
  
  poptContext optCon( poptGetContext( NULL, argc, (const char **)argv,
                                      optionsTable, 0 ) );
  poptSetOtherOptionHelp( optCon, "[FILE]..." );

  try {

    // Parse command-line.
    char c;
    while ( ( c = poptGetNextOpt( optCon ) ) >= 0 );

    // Throw exception on error.
    MMERROR( c >= -1, mimasexception, ,
             poptBadOption( optCon, POPT_BADOPTION_NOALIAS ) << ": "
             << poptStrerror( c ) );
    MMERROR( threshold != -1, mimasexception, ,
             "No value for threshold was specified." );
    MMERROR( fringe >= 0, mimasexception, ,
             "Fringe must be greater or equal to zero." );
    MMERROR( sample >= 1, mimasexception, ,
             "Sampling rate must be greater or equal to one." );

    shared_ptr< image_input< rgba< unsigned char > > > imageInput =
      open_image_input( selectedInput, optCon,
                        gstreamerSHMKey, gstreamerSEMKey,
                        device, channel,
                        width, height, brightness, contrast, shutter, gain,
                        balance );

    x11_display display;

    shared_ptr< image_mesaoutput< unsigned char > > win1, win2;
    
    bool initialised = false;

    while ( true ) {

      image< rgba< unsigned char > > colourImage;
      *imageInput >> colourImage;
      if ( !*imageInput ) break;
      image< unsigned char > img( colourImage );

#ifndef DESPECKLE
      image< bool > edges
        ( edgeDetect( down_sample( image< unsigned char >( img ),
                                   sample, fringe ),
                      threshold ) );
#else
      image< unsigned char > processed;
      despeckleMedian( down_sample( img, sample, fringe ), processed );
      image< bool > edges( edgeDetect( processed, threshold ) );
#endif

      if ( !initialised ) {

        win1 = shared_ptr< image_mesaoutput< unsigned char > >
          ( new image_mesaoutput< unsigned char >
            ( &display, img.getWidth(), img.getHeight() ) );
        win2 = shared_ptr< image_mesaoutput< unsigned char > >
          ( new image_mesaoutput< unsigned char >
            ( &display, edges.getWidth(), edges.getHeight() ) );
        initialised = true;

      };

      *win1 << img;
      *win2 << tobw( edges );

      if ( savePics ) {

        ostringstream outputFileName;
        outputFileName << "edge" << setw( 6 ) << setfill( '0' )
                       << count << resetiosflags( ios::floatfield )
                       << ".ppm";
        
        ofstream ofile( outputFileName.str().c_str(), ios::binary );
        MMERROR( ofile << setformat( "PPM" ) << tobw( edges ),
                 mimasexception, ,
                 "Error writing to file \""
                 << outputFileName.str().c_str() << '\"' );

        count++;
        
      };

      cout << "Number of edge-points was " << numFeatures( edges ) << '.'
           << endl;

    };

  } catch ( exception &e ) {
    cerr << e.what() << endl;
    // Display command-line information.
    poptPrintUsage( optCon, stderr, 0 );
    retVal = 1;
  };

  // Call destructor of command-line parser.
  poptFreeContext( optCon );

  return retVal;
}
