#ifndef __MATCH_H
#define __MATCH_H

#include "instance_data.hh"
#include "mi_pose.hh"

/** Object representing a hypothesis for internal use.
    This object represents a hypothesis about an instance of a class of
    micro-object being found in a scene.

    Each recognition/tracking-algorithm deliveres a set of hypotheses to
    be filtered by the main method \c mi_recognition::recognise.
    @see mi_pose
    @see mi_recognition::recognise
    @author Jan Wedekind (j.wedekind@shu.ac.uk) */
class match
{
public:
  /** Constructor.
      @param _type Class of recognised object.
      @param _pose Hypothetical pose of object.
      @param _fit Value indicating the likelihood of this hypothesis.
      @param _data Internal data of vision-method.
      @see database */
  match( std::string _type, mi_pose _pose, double _fit,
         instance_data_ptr _data = instance_data_ptr() ):
    type(_type), pose(_pose), fit(_fit), data(_data) {}
  /** Get class of recognised object.
      The string must match one of the identifiers in the database of
      micro-objects.
      @return String classifying the micro-object.
      @see database */
  const std::string &get_type(void) const { return type; }
  /** Get pose of instance.
      @return Hypothetical pose of object. */
  mi_pose get_pose(void) const { return pose; }
  /** Get likelihood of hypothesis.
      This is a value, indicating the likelihood of this hypothesis.
      The value is used for filtering out unlikely hypotheses later.
      @return Likelihood of hypothesis. */
  double get_fit(void) const { return fit; }
  /// Get internal data for vision-method.
  instance_data_ptr get_data(void) const { return data; }
protected:
  /// Clasification of object.
  std::string type;
  /// Hypothetical pose.
  mi_pose pose;
  /// Likelihood of hypothesis.
  double fit;
  /// Internal data for vision-method.
  instance_data_ptr data;
};

#endif
