#ifndef __MI_OBJECT_INSTANCE_H
#define __MI_OBJECT_INSTANCE_H

#include <boost/smart_ptr.hpp>
#include <map>
#include <mimas/object.h>
#include <utility>
#include "instance_data.hh"
#include "mi_object_type.hh"
#include "mi_pose.hh"
#include "vision_type.hh"

/** @addtogroup public
    @{ */
/** Hypothesis specifying a micro-object having been found in the scene.
    This object represents a recognition result for a single instance of
    a class of micro-object.

    In contrast to \c match, this is the result <EM>after</EM> filtering
    out unlikely hypotheses.

    This class is part of the public interface.
    @author Jan Wedekind (j.wedekind@shu.ac.uk) */
class mi_object_instance: public mimas::object
{
public:
  /** Constructor.
      @param _type Class of recognised object.
      @param _nr Number of instance for discerning several objects of same
      class.
      @param _timestamp Timestamp of image, on which the recognition was
      performed.
      @param _pose External pose (customized coordinate-system).
      @param type Vision-method, which was employed.
      @param _data Internal data of vision-method. */
  mi_object_instance( const std::string &_type, int _nr,
                      double _timestamp,
                      mi_pose _pose,
                      vision_type type,
                      instance_data_ptr _data ):
    id( _type, _nr ), timestamp( _timestamp ), pose( _pose )
  { internal_data[ type ] = _data; }
  /** Get class of recognised object.
      The string must match one of the identifiers in the database of
      micro-objects.
      @return String classifying the micro-object.
      @see database */
  std::string get_type(void) const { return id.first; }
  /** Get number of instance for discerning several objects of same class.
      @return Number of instance. */
  int get_id(void) const { return id.second; }
  /** Get timestamp of recognition.
      The timestamp of the image, on which the recognition was done, is
      returned. It indicates the time, where the micro-object was at this
      pose.
      @return Timestamp */
  double get_timestamp(void) const { return timestamp; }
  /** Get external pose of micro-object.
      The pose is specifying the transform from the micro-object's
      coordinate-system to the camera's coordinate-system.

      The relation between the object and its coordinate-system is
      customizable with the xml-configuration-file.
      @return Pose of object. */
  const mi_pose &get_pose(void) const { return pose; }
  /// Internal data for vision-algorithms.
  instance_data_ptr data( vision_type type ) { return internal_data[ type ]; }
protected:
  /// Unique identification of the micro-object.
  std::pair< std::string, int > id;
  /// Timestamp of recognition.
  double timestamp;
  /// External pose of micro-object.
  mi_pose pose;
  /// Internal data for vision-algorithms.
  std::map< vision_type, instance_data_ptr > internal_data;
};

/// Pointer to hypothesis.
typedef boost::shared_ptr< mi_object_instance > mi_object_instance_ptr;

///@}

#endif
