#include <cassert>
#include <mimas/linalg.h>
#include "mi_pose.hh"
#include "units.h"

using namespace boost::numeric;
using namespace mimas;

mi_pose::mi_pose( const xml_reference< XalanElement > &configuration )
{
  *this = mi_pose( atof( configuration.selectNode( "rotation/yaw" ).
                         getNodeText().c_str() ) / 180.0 * M_PI,
                   atof( configuration.selectNode( "rotation/pitch" ).
                         getNodeText().c_str() ) / 180.0 * M_PI,
                   atof( configuration.selectNode( "rotation/roll" ).
                         getNodeText().c_str() ) / 180.0 * M_PI,
                   atof( configuration.selectNode( "translation/x" ).
                         getNodeText().c_str() ) * MICROMETER,
                   atof( configuration.selectNode( "translation/y" ).
                         getNodeText().c_str() ) * MICROMETER,
                   atof( configuration.selectNode( "translation/z" ).
                         getNodeText().c_str() ) * MICROMETER );
}

mi_pose::mi_pose( const double w, const double t1, const double t2 ):
  rotation( 2, 2 ), translation( 2 )
{
  double
    cw = cos( w ),
    sw = sin( w );
  rotation( 0, 0 ) =  cw; rotation( 0, 1 ) = -sw;
  rotation( 1, 0 ) =  sw; rotation( 1, 1 ) =  cw;
  translation( 0 ) = t1;
  translation( 1 ) = t2;
}

mi_pose::mi_pose( double yaw, double pitch, double roll,
                  double t1, double t2, double t3 ):
  rotation( 3, 3 ), translation( 3 )
{
  Matrix
    yawMatrix( rotationBlock( 1, 0, 3, yaw ) ),
    pitchMatrix( rotationBlock( 0, 2, 3, pitch ) ),
    rollMatrix( rotationBlock( 2, 1, 3, roll ) );

  // Nested functions won't work in release-mode :-(
  Matrix temp( prod( yawMatrix, pitchMatrix ) );
  rotation = prod( temp, rollMatrix );

  translation[ 0 ] = t1;
  translation[ 1 ] = t2;
  translation[ 2 ] = t3;
}

mi_pose::mi_pose( const Matrix &_rotation, const Vector &_translation ):
  rotation(_rotation), translation(_translation)
{
  assert( rotation.size2() == rotation.size1() );
  assert( translation.size() == rotation.size1() );
}

mi_pose mi_pose::inv(void) const
{
  Matrix rotationInv( mimas::inv< double >( rotation ) );
  return mi_pose( rotationInv, -prod( rotationInv, translation ) );
}

mi_pose mi_pose::operator*( const mi_pose &a ) const
{
  return mi_pose( prod( rotation, a.rotation ),
                  prod( rotation, a.translation ) + translation );
}

mi_pose mi_pose::pose2d(void) const
{
  Matrix rotate( 2, 2 );
  Vector translate( 2 );
  rotate( 0, 0 ) = rotation( 0, 0 );
  rotate( 0, 1 ) = rotation( 0, 1 );
  rotate( 1, 0 ) = rotation( 1, 0 );
  rotate( 1, 1 ) = rotation( 1, 1 );
  translate( 0 ) = translation( 0 );
  translate( 1 ) = translation( 1 );
  return mi_pose( rotate, translate );
}

mi_pose::Matrix mi_pose::rotationBlock( int i1, int i2, int dim, double w )
{
  Matrix retVal( ublas::identity_matrix< double >( dim, dim ) );

  double
    cosW = cos( w ),
    sinW = sin( w );

  retVal( i1, i1 ) =  cosW;
  retVal( i1, i2 ) = -sinW;
  retVal( i2, i1 ) =  sinW;
  retVal( i2, i2 ) =  cosW;

  return retVal;
}

mi_pose mi_pose::scale_2d( double sx, double sy )
{
  Matrix scale( 2, 2 );
  Vector null( 2 ); null[0] = 0.0; null[1] = 0.0;
  scale( 0, 0 ) = sx;
  scale( 0, 1 ) = 0.0;
  scale( 1, 0 ) = 0.0;
  scale( 1, 1 ) = sy;
  return mi_pose( scale, null );
}

mi_pose mi_pose::scale_3d( double sx, double sy, double sz )
{
  Matrix scale( 3, 3 );
  Vector null( 3 ); null[0] = 0.0; null[1] = 0.0;
  scale( 0, 0 ) = sx;
  scale( 0, 1 ) = 0.0;
  scale( 0, 2 ) = 0.0;
  scale( 1, 0 ) = 0.0;
  scale( 1, 1 ) = sy;
  scale( 1, 2 ) = 0.0;
  scale( 2, 0 ) = 0.0;
  scale( 2, 1 ) = 0.0;
  scale( 2, 2 ) = sz;
  return mi_pose( scale, null );
}

mi_pose mi_pose::translate_2d( double tx, double ty )
{
  Matrix identity( ublas::identity_matrix< double >( 2 ) );
  Vector translate( 2 );
  translate( 0 ) = tx;
  translate( 1 ) = ty;
  return mi_pose( identity, translate );
}

mi_pose mi_pose::translate_3d( double tx, double ty, double tz )
{
  Matrix identity( ublas::identity_matrix< double >( 3 ) );
  Vector translate( 3 );
  translate( 0 ) = tx;
  translate( 1 ) = ty;
  translate( 2 ) = tz;
  return mi_pose( identity, translate );
}

mi_pose mi_pose::rotate_2d( double w )
{
  Matrix rotation( 2, 2 );
  double
    cw = cos( w ),
    sw = sin( w );
  rotation( 0, 0 ) =  cw; rotation( 0, 1 ) = -sw;
  rotation( 1, 0 ) =  sw; rotation( 1, 1 ) =  cw;
  Vector null( 2 ); null[0] = 0.0; null[1] = 0.0;
  return mi_pose( rotation, null );
}

mi_pose mi_pose::rotate_3d( double yaw, double pitch, double roll )
{
  return mi_pose( yaw, pitch, roll, 0.0, 0.0, 0.0 );
}

