#ifndef __MI_POSE_H
#define __MI_POSE_H

#include <boost/shared_ptr.hpp>
#include <boost/numeric/ublas/vector.hpp>
#include <boost/numeric/ublas/matrix.hpp>
#include <mimas/object.h>
#include <mimas/xml_reference.h>

/** @addtogroup public
    @{ */
/** Pose of a micro-object.
    The pose of a micro-object is defined by the affine transformation
    between it's coordinate-system and the superior coordinate-system
    (the camera's coordinate-system):
    \f[
    \vec{x}^\prime=A\,\vec{x}+\vec{t}
    \f]
    where \f$\vec{x}=\vec{OX}\f$ specifies the position of the point
    \f$X\f$ in the coordinate-system of the micro-object.

    This class is part of the public interface.
    @author Jan Wedekind (j.wedekind@shu.ac.uk) */
class mi_pose: public mimas::object
{
public:
  /// Type-definition for convenience.
  typedef boost::numeric::ublas::matrix< double > Matrix;
  /// Type-definition for convenience.
  typedef boost::numeric::ublas::vector< double > Vector;
  /** Construct by loading from XML-DOM.
      Initialise 6-DOF pose from XML-data.
      @param configuration root-node of XML-subtree. */
  mi_pose( const mimas::xml_reference< XalanElement > &configuration );
  /** Constructor for identity transform.
      Create an identity transformation with \f$A:=I\f$
      (\f$I_{i,j}:=\delta_{i,j}\f$, \f$\delta_{i,j}\f$ is Kronecker-function)
      and \f$\vec{t}=\vec{o}\f$.
      @param dim Dimension of coordinate systems (default is 3). */
  mi_pose( int dim = 3 ):
    rotation( boost::numeric::ublas::identity_matrix< double >( dim ) ),
    translation( dim ) {}
  /** Constructor for a two-dimensional transformation.
      Counter-clockwise rotation followed by a translation.
      \verbatim
           [  cos(y) -sin(y) ] 
      A := [  sin(y)  cos(y) ]
      \endverbatim
      \f$\vec{t}=(t_1\ t_2)^\top\f$.
      @param w Angle of rotation (in radian).
      @param t1 First element of translation-vector.
      @param t2 Second element of translation-vector. */
  mi_pose( const double w, const double t1, const double t2 );
  /** Constructor for a three-dimensional transformation.
      Note that the rotation around the z-axis (yaw-angle) is clockwise
      by convention.
      \verbatim
           [  cos(y)  sin(y) 0 ] 
      Y := [ -sin(y)  cos(y) 0 ]
           [    0       0    1 ]

           [  cos(p) 0 -sin(p) ] 
      P := [    0    1    0    ]
           [  sin(p) 0  cos(p) ]

           [ 1    0       0    ]
      R := [ 0  cos(r)  sin(r) ]
           [ 0 -sin(r)  cos(r) ]

      A := Y * P * R
      \endverbatim
      @param yaw Rotation around z-axis (radian).
      @param pitch Pitch angle (radian).
      @param roll Roll angle (radian).
      @param t1 First element of translation-vector.
      @param t2 Second element of translation-vector.
      @param t3 Third element of translation-vector. */
  mi_pose( double yaw, double pitch, double roll,
           double t1, double t2, double t3 );
  /** Construct from given matrix and vector.
      @param _rotation Matrix \f$A\f$ of transformation.
      @param _translation Vector \f$\vec{t}\f$ of transformation. */
  mi_pose( const Matrix &_rotation, const Vector &_translation );
  /** Copy constructor.
      @param _pose Object to copy from. */
  mi_pose( const mi_pose &_pose ):
    rotation( _pose.rotation ), translation( _pose.translation ) {}
  /** Compute inverse of transform.
      Computes the transform \f$(A^\prime,\vec{t}^\prime)\f$, such that
      \f[
      \forall\vec{x}:\,\vec{x}=A^\prime\,(A\,\vec{x}+\vec{t})+\vec{t}^\prime
      \f]
      holds.
      @return The inverse transformation. */
  mi_pose inv(void) const;
  /** Perform transformation.
      Computes
      \f[
      \vec{x}^\prime=A\,\vec{x}+\vec{t}
      \f]
      @param x The vector \f$\vec{x}\f$
      @return The vector \f$\vec{x}^\prime\f$. */
  Vector map( const Vector &x ) const
    { return prod( rotation, x ) + translation; }
  /** Concatenate two transformations.
      @param a right argument of multiplication. */
  mi_pose operator*( const mi_pose &a ) const;
  /** Get const-reference to rotational matrix.
      @return The matrix \f$A\f$. */
  const Matrix &getRotation(void) const { return rotation; }
  /** Get const-reference to translational vector.
      @return The vector \f$\vec{t}\f$. */
  const Vector &getTranslation(void) const { return translation; }
  /** Get reference to rotational matrix.
      @return The matrix \f$A\f$. */
  Matrix &getRotation(void) { return rotation; }
  /** Get reference to translational vector.
      @return The vector \f$\vec{t}\f$. */
  Vector &getTranslation(void) { return translation; }
  /// Extract transformation for first two coordinates.
  mi_pose pose2d(void) const;
  /** Scaling 2-d transform.
      @param sx scale-factor for x-coordinates.
      @param sy scale-factor for y-coordinates. */
  static mi_pose scale_2d( double sx, double sy );
  /** Scaling 2-d transform.
      @param sx scale-factor for x-coordinates.
      @param sy scale-factor for y-coordinates.
      @param sz scale-factor for z-coordinates. */
  static mi_pose scale_3d( double sx, double sy, double sz );
  /** Translating 2-d transform.
      @param tx translation for x-coordinates.
      @param ty translation for y-coordinates. */
  static mi_pose translate_2d( double tx, double ty );
  /** Translating 3-d transform.
      @param tx translation for x-coordinates.
      @param ty translation for y-coordinates.
      @param tz translation for y-coordinates. */
  static mi_pose translate_3d( double tx, double ty, double tz );
  /** Rotating 2-d transform.
      @param w angle. */
  static mi_pose rotate_2d( double w );
  /** 3-d rotation.
      @param yaw Rotation around z-axis (radian).
      @param pitch Pitch angle (radian).
      @param roll Roll angle (radian). */
  static mi_pose rotate_3d( double yaw, double pitch, double roll );
protected:
  /** Generate matrix with a single rotation block.
      @param i1 major index of rotation-block.
      @param i2 minor index of rotation-block.
      @param dim Dimensionality of matrix.
      @param w Angle of rotation. */
  Matrix rotationBlock( int i1, int i2, int dim, double w );
  /// Matrix of transformation.
  Matrix rotation;
  /// Vector of transformation.
  Vector translation;
};

/// Pointer to pose-object.
typedef boost::shared_ptr< mi_pose > mi_pose_ptr;

///@}

#endif
