#include "plugins/hough-tracking/hough_tracking_plugin.hh"
#include <cmath>
#include <mimas/xml_node_reference_list.h>
#include <set>
#include "match.hh"
#include "plugins/geometric-voting/geometric_voting_plugin.hh"
#include "plugins/human-vision/human_vision_plugin.hh"
#include "plugins/maxima-extraction/maxima_extraction_plugin.hh"
#include "plugins/bounded-hough/bounded_hough_plugin.hh"
#include "mi_recognition.hh"
#include "units.h"

using namespace mimas;
using namespace std;

mi_recognition::mi_recognition( const mimas::xml_reference< XalanElement >
                                &configuration, bool _verbose ):
  verbose(_verbose)
{
  // Load calibration data.
  calib =
    calibration_ptr( new calibration
                     ( dynamic_pointer_cast< XalanElement >
                       ( configuration.selectNode( "calibration" ) ) ) );

  // Read offset-list.
  xml_node_reference_list
    objects( configuration.selectNodes( "database/object-type" ) );

  // Read pose from xml-file.
  for ( int i=0; i<objects.getLength(); i++ ) {
    xml_reference< XalanElement > object
      ( dynamic_pointer_cast< XalanElement >( objects.item( i ) ) );
    mi_object_type_ptr object_type( new mi_object_type( object ) );
    string name = object.selectNode( "@name" ).getNodeText();
    MMERROR( object_types.find( name ) == object_types.end(), mimasexception, ,
             "Duplicate micro-object type \"" << name << "\"." );
    object_types[ name ] = object_type;
  };

  // Initialise recognition by geometric-voting.
  if ( configuration.existNode( "methods/geometric-voting" ) )
    methods[ "geometric-voting" ] =
      recognition_method_ptr( new geometric_voting_plugin( calib,
                                                           configuration,
                                                           verbose ) );

  // Initialise stencil-tracking.
  if ( configuration.existNode( "methods/hough-tracking" ) )
    methods[ "hough-tracking" ] =
      recognition_method_ptr( new hough_tracking_plugin( calib,
                                                         configuration,
                                                         verbose ) );

  // Add wrapper for "human vision" :-(
  if ( configuration.existNode( "methods/human-vision" ) )
    methods[ "human-vision" ] =
      recognition_method_ptr( new human_vision_plugin( calib, configuration,
                                                       this, verbose ) );

  // Add maxima-extraction (custom method for finding solder-sphere on black
  // background).
  if ( configuration.existNode( "methods/maxima-extraction" ) )
    methods[ "maxima-extraction" ] =
      recognition_method_ptr( new maxima_extraction_plugin( calib,
                                                            configuration,
                                                            verbose ) );

  // Initialise bounded hough transform.
  if ( configuration.existNode( "methods/bounded-hough" ) )
    methods[ "bounded-hough" ] =
      recognition_method_ptr( new bounded_hough_plugin( calib,
                                                        configuration,
                                                        verbose ) );

}

void mi_recognition::remove_objects( const string &method,
                                     mi_scene_ptr scene )
  const
{
  vector< mi_object_instance_ptr > *objects = &scene->get_objects();
  int i=0;

  while ( i < (signed)objects->size() ) {
    string type = (*objects)[i]->get_type();
    assert( object_types.find( type ) != object_types.end() );
    mi_object_type_ptr t
      ( object_types.find( (*objects)[i]->get_type() )->second );
    if ( method == t->get_recognition_method() ||
         method == t->get_tracking_method() )
      objects->erase( objects->begin() + i );
    else
      i++;
  };
}

void mi_recognition::reschedule( const string &method,
                                 mi_scene_ptr scene ) const
{
  set< string > types;

  vector< mi_object_instance_ptr > *objects = &scene->get_objects();
  int i = 0;

  while ( i < (signed)objects->size() ) {
    string type = (*objects)[i]->get_type();
    assert( object_types.find( type ) != object_types.end() );
    mi_object_type_ptr t
      ( object_types.find( type )->second );
    if ( method == t->get_tracking_method() ||
         method == t->get_recognition_method() ) {
      scene->get_jobs()[ type ] = 1;
      objects->erase( objects->begin() + i );
    } else
      i++;
  };  
}

void mi_recognition::take_best_matches( std::vector< match > &matches,
                                        double timestamp,
                                        mi_scene_ptr scene,
                                        vision_type method ) const
{
  // Filtering of hypotheses/matches is done by just taking the best match
  // for each class.
  map< string, match * > maxElement;
  for ( vector< match >::iterator i=matches.begin();
          i != matches.end(); i++ ) {
    map< string, match * >::iterator entry
      ( maxElement.find( i->get_type() ) );
    if ( entry != maxElement.end() ) {
      if ( entry->second->get_fit() < i->get_fit() )
        entry->second =&(*i); 
    } else
      maxElement[ i->get_type() ] = &(*i);
  };

  // Insert the best matches into the resulting list.
  for ( map< string, match * >::iterator i = maxElement.begin();
        i != maxElement.end(); i++ ) {

    scene->get_objects().push_back
      ( mi_object_instance_ptr
        ( new mi_object_instance
          ( i->second->get_type(), 0, timestamp, i->second->get_pose(),
            method, i->second->get_data() ) ) );

  };
}

void mi_recognition::recognise( const image< unsigned char > &image,
                                double timestamp, mi_scene_ptr scene ) const
{

  // Check image-size.
  MMERROR( image.getWidth() == calib->get_width() &&
           image.getHeight() == calib->get_height(), mimasexception, ,
           "Image was of size " << image.getWidth() << "x"
           << image.getHeight() << ", but should be of size "
           << calib->get_width() << "x" << calib->get_height()
           << " as stated in the configuration." );

  if ( verbose )
    cout << "Computer-vision for image with timestamp " << timestamp << endl;

  // Make list of tracking-methods, which have to be executed.
  set< string > active_tracking_methods;
  for ( vector< mi_object_instance_ptr >::const_iterator i =
          scene->get_objects().begin(); i != scene->get_objects().end();
        i++ ) {
    assert( object_types.find( (*i)->get_type() ) != object_types.end() );
    mi_object_type_ptr object_type
      ( object_types.find( (*i)->get_type() )->second );
    active_tracking_methods.insert( object_type->get_tracking_method() );
  };

  // Apply all active tracking-methods.
  for ( set< string >::const_iterator i = active_tracking_methods.begin();
        i != active_tracking_methods.end(); i++ ) {

    if ( methods.find( *i ) != methods.end() ) {

      // Apply tracking-method.
      recognition_method_ptr tracking_method = methods.find( *i )->second;
      vector< match > result =
        tracking_method->recognise( image, timestamp, scene );

      // At the moment, finding a correspondence between a hypotheses and a
      // previously recognised object-instances is easy, as there is only
      // one instance per class at the most.
      // Clear object-list.
      remove_objects( *i, scene );
      // Add acquired positions to empty list.
      take_best_matches( result, timestamp, scene,
                         tracking_method->get_type() );

    } else {

      // Could not apply method. The micro-objects have to be scheduled for
      // recognition.
      reschedule( *i, scene );

    };
  };
  
  // Make list of recognition-methods, which have to be executed.
  set< string > active_recognition_methods;
  for ( map< string, int >::const_iterator i =
          scene->get_jobs().begin(); i != scene->get_jobs().end();
        i++ ) {
    assert( object_types.find( i->first ) != object_types.end() );
    mi_object_type_ptr object_type
      ( object_types.find( i->first )->second );
    active_recognition_methods.insert( object_type->get_recognition_method() );
  };

  // Apply all active recognition-methods.
  for ( set< string >::const_iterator i = active_recognition_methods.begin();
        i != active_recognition_methods.end(); i++ )

    if ( methods.find( *i ) != methods.end() ) {
    
      // Apply recognition-method.
      recognition_method_ptr recognition_method = methods.find( *i )->second;
      vector< match > result =
        recognition_method->recognise( image, timestamp, scene );
    
      // Add acquired positions to object-list.
      take_best_matches( result, timestamp, scene,
                         recognition_method->get_type() );
    
    } else
      MMERROR( false, mimasexception, ,
               "Vision-method \"" << *i
               << "\" is required, but wasn't initialised." );

  // Clear job-list.
  scene->get_jobs().clear();

  if ( verbose ) cout << endl;
}

bool mi_recognition::human_vision( const string &,
                                   const image< unsigned char > &, double,
                                   double *, int *, int * ) const
{
  return false;
}
