#ifndef __MI_RECOGNITION_H
#define __MI_RECOGNITION_H

#include <map>
#include <mimas/image.h>
#include <mimas/object.h>
#include <mimas/xml_reference.h>
#include <vector>
#include "calibration.hh"
#include "mi_object_type.hh"
#include "mi_scene.hh"
#include "recognition_method.hh"

/** @defgroup public Public interface
    Public interface for micron-vision. */
///@{
/** Recognition class.
    The recognition class hides the use of different recognition- and
    tracking-methods from the calling program.

    This class is part of the public interface.
    @author Jan Wedekind (j.wedekind@shu.ac.uk) */
class mi_recognition: public mimas::object
{
public:
  /** Constructor.
      Initialise the recognition algorithms and the database of
      micro-objects.
      @param configuration XML-tree with configuration. 
      @param _verbose Set to \c true, if you want to get verbose information
      from the vision-algorithms on stdout. */
  mi_recognition( const mimas::xml_reference< XalanElement >
                  &configuration, bool _verbose );
  /** Perform recognition/tracking.
      The micro-objects known to the \c scene object are being tracked.
      Any new recognition-jobs, being specified in \c scene, are attempted.
      The model contained in \c scene is updated accordingly then.
      @param image Image to recognise/track micro-objects in.
      @param timestamp Time of image-capture.
      @param scene Object encapsulating the state of vision.
      @see units */
  void recognise( const mimas::image< unsigned char > &image,
                  double timestamp, mi_scene_ptr scene ) const;
  /** Access calibration object.
      Get access to parameters of camera-calibration
      @return Pointer to camera-calibration object. */
  calibration_ptr get_calibration(void) { return calib; }
  /** Access list with types of micro-objects.
      A map is returned, which maps strings on pointers to
      mi_object_type objects.
      @return Reference to the map.
      @see mi_object_type */
  std::map< std::string, mi_object_type_ptr > &get_object_types(void)
    { return object_types; }
  /* Add user-defined method (being used for "human recognition").
      @param name Name of recognition-method.
      @param method The recognition-object. */
  /* void add_method( const std::string &name, recognition_method_ptr method )
     { methods[ name ] = method; } */
  /** Human vision.
      Recognise objects using a human.
      If you want to recognise an object with human-vision, you need to
      override this method and modify the configuration-file
      (see f.e. gripper.xml).
      @param type Type of object to be recognised.
      @param image Image to recognise object in.
      @param timestamp Timestamp of image.
      @param angle Angle of object.
      @param x x-coordinate of object.
      @param y y-coordinate of object.
      @return \c true, if recognition was successful.
      @see human_vision_plugin */
  virtual bool human_vision( const std::string &type,
                             const mimas::image< unsigned char > &img,
                             double timestamp, double *angle,
                             int *x, int *y ) const;
protected:
  /** Remove some objects from the tracking-list.
      All entries of micro-objects being tracked with a specific vision-method
      are removed.
      @param method Tracking-method responsible for this object.
      @param scene Scene-object with tracking-list. */
  void remove_objects( const std::string &method,
                       mi_scene_ptr scene ) const;
  /** Remove some objects from the tracking-list.
      All entries of micro-objects being tracked/recognised with a specific
      vision-method are rescheduled for recognition and removed from the
      tracking-list.
      @param method Tracking-method responsible for this object.
      @param scene Scene-object with tracking-list and job-list. */
  void reschedule( const std::string &method,
                   mi_scene_ptr scene ) const;
  /** Take best recognition-results.
      @param matches List of matches.
      @param timestamp Timestamp of image.
      @param scene Scene-object to be updated with new recognition-results.
      @param method Type of vision-method (tracking or recognition). */
  void take_best_matches( std::vector< match > &matches,
                          double timestamp,
                          mi_scene_ptr scene,
                          vision_type method ) const;
  /// Calibration object.
  calibration_ptr calib;
  /// Initial recognition.
  std::map< std::string, recognition_method_ptr > methods;
  /// General information about the different type of micro-objects.
  std::map< std::string, mi_object_type_ptr > object_types;
  ///
  bool verbose;
};

///
typedef boost::shared_ptr< mi_recognition > mi_recognition_ptr;

///@}

#endif
