#ifndef __MI_SCENE_H
#define __MI_SCENE_H

#include <boost/smart_ptr.hpp>
#include <map>
#include <mimas/object.h>
#include <vector>
#include "mi_object_instance.hh"

/** @addtogroup public
    @{ */
/** A scene keeps track of the recognised micro-objects.
    This object is used for exchanging data between the vision-system and
    the calling program. \c mi_recognition objects must store all information
    in this object, which is required for the later recognition-steps.

    This class is part of the public interface.
    @see mi_recognition
    @see mi_recognition::recognise
    @author Jan Wedekind (j.wedekind@shu.ac.uk) */
class mi_scene: public mimas::object
{
public:
  /** Constructor.
      Constructs an empty scene. */
  mi_scene(void) {}
  /** Gain access to list of micro-objects.
      This list of objects contains the micro-objects, which are being tracked
      at the moment. If a object shouldn't be tracked any more, it has to be
      removed by the calling program.
      @return List of micro-objects.
      @see mi_object_instance */
  std::vector< mi_object_instance_ptr > &get_objects(void) { return objects; }
  /// Same as above (but const).
  const std::vector< mi_object_instance_ptr > &get_objects(void) const
  { return objects; }
  /** Add a recognition task.
      A job to recognise a new object is being specified. The recognition-job
      will only be attempted once. Afterwards it has to be added again.

      If there are two conflicting specifications for the same type of
      micro-object, the first one will be overriden.
      @param maxNum Maximum number of instances to be expected.
      @param name Name of type of micro-object.
      @see mi_recognition::object_types
      @see mi_recognition::recognise */
  void add_job( const std::string &name, int maxNum );
  /// Access job-list.
  std::map< std::string, int > &get_jobs(void) { return jobs; }
protected:
  /** List of micro-objects.
      This is the list of micro-object being tracked at the moment.
      @see mi_object_instance */
  std::vector< mi_object_instance_ptr > objects;
  /** List of recognition-jobs.
      This is the list of recognition-jobs to be attempted on the next
      image. */
  std::map< std::string, int > jobs;
};

/// Pointer to object of type mi_scene.
typedef boost::shared_ptr< mi_scene > mi_scene_ptr;

///@}

#endif
