// Human vision only works with qt-output at the moment
// #define HUMANVISION

// Grab only first frame and run it 1000 times.
// #define TESTSINGLEFRAME
#define TESTCOUNT 200
#define TESTSKIP 10

#ifndef HUMANVISION
// Comment it out, if it doesn't work.
#define USEMESAOUTPUT
#endif

#include "micron_vision_config.h"
#include <boost/numeric/ublas/vector.hpp>
#include <boost/smart_ptr.hpp>
#include <cmath>
#include <exception>
#include <fstream>
#include <GL/glut.h>
#include <iomanip>
#include <iostream>
#include <libgen.h>
#include <mimas/image_fileoutput.h>
#include <mimas/x11_window.h>
#include <mimas/mesa_offscreen.h>
#include <mimas/mesa_onscreen.h>
#include <mimas/timer.h>
#include <mimas/xml_document.h>
#include <popt.h>
#ifdef HUMANVISION
#include <qlabel.h>
#include <qimage.h>
#endif
#include <sstream>
#include <XPath/XPathInit.hpp>
#include <framework/MemBufInputSource.hpp>
#include <util/PlatformUtils.hpp>
#include "mi_recognition.hh"
#include "tools.hh"
#include "units.h"

using namespace boost;
using namespace boost::numeric;
using namespace mimas;
using namespace std;

template< typename T >
class image_micronoutput: public x11_window
{
 public:
  image_micronoutput( x11_display_ptr _display,
                      int _width, int _height ):
    x11_window( get_pointer( _display ), _width, _height ),
    timestamp(0.0) {}
  void setInfo( mi_scene_ptr _scene,
                double _timestamp,
                int _processed,
                calibration_ptr _calib )
  {
    scene = _scene;
    timestamp = _timestamp;
    processed = _processed;
    calib = _calib;
  }
  void setImage( const image< rgba< unsigned char > > &_img ) { img = _img; repaint(); }
protected:
  virtual void paint_event(void) throw (mimasexception);
  mi_scene_ptr scene;
  calibration_ptr calib;
  double timestamp;
  int processed;
  image< rgba< unsigned char > > img;
};

#ifdef HUMANVISION

class HumanVisionWidget: public QLabel
{
 public:
  HumanVisionWidget( QWidget *parent):
    QLabel( parent ), recognized(false) {}
  bool isRecognized(void) const { return recognized; }
  int get_x(void) const { return x; }
  int get_y(void) const { return y; }
 protected:
  virtual void mousePressEvent( QMouseEvent *e ) {
    recognized = true;
    mouseMoveEvent( e );
  }
  virtual void mouseMoveEvent( QMouseEvent *e ) {
    x = e->x();
    y = e->y();
  };
  virtual void mouseReleaseEvent( QMouseEvent *e ) {
    recognized = false;
  }
  bool recognized;
  int x;
  int y;
};

static QPixmap image_to_qpixmap
 ( const image< rgba< unsigned char > > &img )
{
  QImage qImg( img.getWidth(), img.getHeight(), 32 );
  assert( sizeof( rgba< unsigned char > ) == 4 );

  // Line-wise copying required, as QImage does 32 bit memory alignment!
  // Direct access to image-memory!
  for ( int y=0; y<img.getHeight(); y++ )
    copy_n( img.rawData()[y].begin(), img.getWidth(),
            (rgba< unsigned char > *)qImg.scanLine( y ) );
  
  QPixmap retVal;
  // Direct access to image-memory!
  // retVal.convertFromImage( qImg.swapRGB() );
  retVal.convertFromImage( qImg );
  return retVal;
}

/// Recognition-class extended with 
class ex_recognition: public mi_recognition
{
 public:
  ex_recognition( const mimas::xml_reference< XalanElement >
                  &configuration, const string &_test_type,
                  bool _verbose );
  virtual ~ex_recognition(void)
  { if ( win ) delete win; }
  virtual bool human_vision( const std::string &type,
                             const mimas::image< unsigned char > &img,
                             double timestamp, double *angle,
                             int *x, int *y ) const {
    if ( win ) {
      win->setPixmap( QPixmap( image_to_qpixmap( img ) ) );
      win->show();
      *x = win->get_x();
      *y = win->get_y();
      *angle = 0.0;// radian
      return ( win->isRecognized() && type == test_type );
    } else
      return false;
  };
 protected:
  string test_type;
  HumanVisionWidget *win;
};

ex_recognition::ex_recognition
( const mimas::xml_reference< XalanElement > &configuration,
  const string &_test_type, bool _verbose ):
  mi_recognition( configuration, _verbose ), test_type( _test_type ), win(NULL)
{
  if ( test_type != "" ) {
    win = new HumanVisionWidget( NULL );
    win->setBackgroundMode( Qt::NoBackground );
  };
}

#endif

void displayCopyright( mesa_base &m, int width )
{
  const rgba< unsigned char > green( 0, 255, 0 );
  glPushMatrix();
  m.setColor( green );
  glLineWidth( 1.0 );
  glTranslated( width / 30.0, width / 30.0, 0 );
  glScaled( width / 6000.0, -width / 6000.0, 1.0 );
  m.text( "Mimas 2005 (C) MMVL, Sheffield Hallam University" );
  glPopMatrix();
}

string header( mi_scene_ptr scene, double timestamp, int processed )
{
  // Create text with position information.
  ostringstream info;
  info << scene->get_objects().size() << " object(s), timestamp = "
       << fixed << setw( 8 ) << setprecision( 2 )
       << ( timestamp / SECOND ) << " s, "
       << resetiosflags( ios::floatfield )
       << processed << "th frame";
  return info.str();
}

double rotation( mi_object_instance_ptr instance )
{
  // Get coordinate-transformation of micro-object (4-DOF).
  const mi_pose *pose( &instance->get_pose() );

  // Extract rotation-angle from matrix (degrees).
  double retVal = atan2( pose->getRotation()( 1, 0 ),
                         pose->getRotation()( 0, 0 ) ) * 180.0 / M_PI;

  return retVal;
}

string instanceInfo( mi_object_instance_ptr instance )
{
  // Get coordinate-transformation of micro-object (4-DOF).
  const mi_pose *pose( &instance->get_pose() );
  
  // Get translation-vector.
  const ublas::vector< double > *translation( &pose->getTranslation() );

  ostringstream info;
  info << setw( 8 ) << instance->get_type() << " at "
       << "( " << fixed << setw( 8 ) << setprecision( 2 )
       << ( (*translation)(0) / MICROMETER ) << " um, "
       << fixed << setw( 8 ) << setprecision( 2 )
       << ( (*translation)(1) / MICROMETER ) << " um, "
       << fixed << setw( 8 ) << setprecision( 2 )
       << ( (*translation)(2) / MICROMETER ) << " um ), angle is "
       << fixed << setw( 6 ) << setprecision( 1 )
       << rotation( instance ) << resetiosflags( ios::floatfield )
       << " deg.";
  return info.str();
}

void displayInfo( mesa_base &m, mi_scene_ptr scene,
                  calibration_ptr calib, double timestamp, int processed,
                  int width )
{
  m.ortho2D();

  glDisable( GL_LINE_SMOOTH );

  // Display copyright-message.
  displayCopyright( m, width );
  
  // Change color to red.
  const rgba< unsigned char > red( 255, 0, 0 );
  m.setColor( red );
 
  // Print general information.
  glPushMatrix(); {
    glTranslated( width / 30.0, width / 13, 0 );
    glScaled( width / 7500.0, -width / 7500.0, 1.0 );
    m.text( header( scene, timestamp, processed ) );
  }; glPopMatrix();
  
  // Get the size of projected pixels.
  double pixelSize = calib->get_pixel_size();
  
  // Loop for displaying recognised/tracked objects.
  for ( int i=0; i<min( (signed)scene->get_objects().size(), 100 );
        i++ ) {

    mi_object_instance_ptr instance( scene->get_objects()[i] );

    // Get coordinate-transformation of micro-object (4-DOF).
    const mi_pose *pose( &instance->get_pose() );
    // Get translation-vector.
    const ublas::vector< double > *translation( &pose->getTranslation() );
    
    glEnable( GL_LINE_SMOOTH );
    glHint( GL_LINE_SMOOTH_HINT, GL_DONT_CARE );
    glLineWidth( 1.5 );

    // Save matrix on stack.
    glPushMatrix(); {
              
      // Center coordinate system in middle of image.
      glTranslated( calib->get_middle_x(),
                    calib->get_middle_y(),
                    0 );
      // Scale according to camera-resolution.
      glScalef( 1.0 / pixelSize, 1.0 / pixelSize, 1.0 );

      // Tell Mesa about the micro-object's pose.
      glTranslated( (*translation)(0), (*translation)(1), 0 );
      glRotatef( rotation( instance ), 0.0, 0.0, 1.0 );

      // Start drawing single lines.
      glBegin( GL_LINES ); {
        // Draw x-axis.
        glVertex2f( -10 * pixelSize, 0 );
        glVertex2f( +40 * pixelSize, 0 );
        glVertex2f( +35 * pixelSize, -5 * pixelSize );
        glVertex2f( +40 * pixelSize, 0 );
        glVertex2f( +35 * pixelSize, +5 * pixelSize );
        glVertex2f( +40 * pixelSize, 0 );
        // Draw y-axis.
        glVertex2f( 0, -10 * pixelSize );
        glVertex2f( 0, +20 * pixelSize );
        glVertex2f( -5 * pixelSize, +15 * pixelSize );
        glVertex2f( 0, +20 * pixelSize );
        glVertex2f( +5 * pixelSize, +15 * pixelSize );
        glVertex2f( 0, +20 * pixelSize );
      }; glEnd();
              
      // Label x-axis.
      glPushMatrix(); {
        glTranslated( +40 * pixelSize, -10 * pixelSize, 0 );
        glScaled( 0.1 * pixelSize, -0.1 * pixelSize, 1.0 );
        m.text( "x" );
      }; glPopMatrix();

      // Label y-axis.
      glPushMatrix(); {
        glTranslated( -20 * pixelSize, +20 * pixelSize, 0 );
        glScaled( 0.1 * pixelSize, -0.1 * pixelSize, 1.0 );
        m.text( "y" );
      }; glPopMatrix();

    }; glPopMatrix();

    glDisable( GL_LINE_SMOOTH );
    glLineWidth( 1.0 );
    // Print position information.
    glPushMatrix(); {
      glTranslated( width / 30, width / 13 +
                    width / 37.0 * ( i + 1 ), 0 );
      glScaled( width / 8500.0, -width / 8500.0, 1.0 );
      m.text( instanceInfo( instance ) );
    }; glPopMatrix();

  };
}


template<>
void image_micronoutput< unsigned char >::paint_event(void)
  throw (mimasexception)
{
  assert( scene != NULL );
  mesa_onscreen m( this );
  // glClearColor( 0.0, 0.0, 0.0, 0.0 );
  // glClear( GL_ALL_ATTRIB_BITS );
  glPixelStorei( GL_UNPACK_ALIGNMENT, 1 );
  glRasterPos2i( -1, 1 );
  glPixelZoom( 1.0, -1.0 );
  glDrawPixels( get_width(), get_height(), GL_LUMINANCE,
		GL_UNSIGNED_BYTE, img.rawData().data() );

  displayInfo( m, scene, calib, timestamp, processed,
               img.getWidth() );
}

template<>
void image_micronoutput< rgba< unsigned char > >::paint_event(void)
  throw (mimasexception)
{
  assert( scene != NULL );
  mesa_onscreen m( this );
  // glClearColor( 0.0, 0.0, 0.0, 0.0 );
  // glClear( GL_ALL_ATTRIB_BITS );
  glPixelStorei( GL_UNPACK_ALIGNMENT, 1 );
  glRasterPos2i( -1, 1 );
  glPixelZoom( 1.0, -1.0 );
  glDrawPixels( get_width(), get_height(), GL_BGRA,
		GL_UNSIGNED_BYTE, img.rawData().data() );

  displayInfo( m, scene, calib, timestamp, processed,
               img.getWidth() );
}

// Test program for the vision-system for MiCRoN.
int main( int argc, char *argv[] )
{
  glutInit( &argc, argv );
  // Start measuring of time.
  timer initialisation;
  
  // Configure with default values.
  int
    retVal = 0,
    selectedInput = 0,
    gstreamerSHMKey = 333,
    gstreamerSEMKey = 333,
    channel = 0,
    frames = 0,
    frameDrop = 0,
    savePics = 0,
    waitForUser = 0,
    x11Output = 1,
    verbose = 0,
    brightness = 32768,
    contrast = 32768,
    shutter = -1,
    gain = -1,
    balance = -1;

  const char
    *configFileName = "configuration.xml",
    *configSchemaFileName = "configuration.xsd",
    *device = NULL;
#ifdef HUMANVISION
  const char *humanVisionTest = "";
#endif

  // Set up table of command-line options.
  struct poptOption optionsTable[] = {
    { "config", 'c', POPT_ARG_STRING | POPT_ARGFLAG_SHOW_DEFAULT,
      &configFileName, 0, "Specify alternate configuration file",
      "xml-file-name" },
    { "schema", 's', POPT_ARG_STRING | POPT_ARGFLAG_SHOW_DEFAULT,
      &configSchemaFileName, 0, "Specify alternate location of schema file",
      "xsd-file-name" },
    { "gstreamer", 'g', POPT_ARG_VAL, &selectedInput, 1,
      "Select gstreamer input for reading images", 0 },
    { "shm", '\0', POPT_ARG_INT | POPT_ARGFLAG_SHOW_DEFAULT, &gstreamerSHMKey,
      0, "shared memory key for gstreamer input", "integer" },
    { "sem", '\0', POPT_ARG_INT | POPT_ARGFLAG_SHOW_DEFAULT, &gstreamerSEMKey,
      0, "semaphore key for gstreamer input", "integer" },
    { "v4l", 'v', POPT_ARG_VAL, &selectedInput, 2,
      "Select video4linux for reading images", 0 },
    { "brightness", '\0', POPT_ARG_INT | POPT_ARGFLAG_SHOW_DEFAULT,
      &brightness, 0, "Brightness for video4linux-input [0..65535]",
      "brightness" },
    { "contrast", '\0', POPT_ARG_INT | POPT_ARGFLAG_SHOW_DEFAULT,
      &contrast, 0, "Contrast for video4linux-input [0..65535]", "contrast" },
#ifdef HAVE_LIBDC1394
    { "video1394", '\0', POPT_ARG_VAL, &selectedInput, 3,
      "Select firewire digital camera for reading images", 0 },
    { "shutter", '\0', POPT_ARG_INT | POPT_ARGFLAG_SHOW_DEFAULT,
      &shutter, 0, "Shutter for video1394-input", "shutter" },
    { "gain", '\0', POPT_ARG_INT | POPT_ARGFLAG_SHOW_DEFAULT,
      &gain, 0, "Gain for video1394-input", "gain" },
    { "balance", '\0', POPT_ARG_INT | POPT_ARGFLAG_SHOW_DEFAULT,
      &balance, 0, "White balance for video1394-input", "balance" },
#endif
    { "device", '\0', POPT_ARG_STRING,
      &device, 0, "Specify video4linux-/video1394-device", "device-name" },
    { "channel", '\0', POPT_ARG_INT | POPT_ARGFLAG_SHOW_DEFAULT,
      &channel, 0, "Specify video4linux-channel/video1394-node",
      "channel-number" },
    { "fps", '\0', POPT_ARG_INT, &frames, 0,
      "use number of image divided by given frame-rate for timestamp "
      "(default: use time elapsed)", "integer" },
    { "framedrop", 'd', POPT_ARG_VAL, &frameDrop, 1,
      "Drop frames, if algorithm is slower than given frame-rate", 0 },
    { "savepics", '\0', POPT_ARG_VAL, &savePics, 1,
      "Save pictures for debugging-purposes (debug*.ppm)", 0 },
    { "wait", 'w', POPT_ARG_VAL, &waitForUser, 1,
      "Pause after initialisation", 0 },
    { "verbose", '\0', POPT_ARG_VAL, &verbose, 1,
      "Get verbose output from vision-methods", 0 },
    { "disablex11", 'x', POPT_ARG_VAL, &x11Output, 0,
      "Disable graphical output", 0 },
    { "enablex11", 'X', POPT_ARG_VAL, &x11Output, 1,
      "Enable graphical output (default)", 0 },
#ifdef HUMANVISION
    { "human", '\0', POPT_ARG_STRING, &humanVisionTest, 0,
      "Test human-vision API at given time.", "type of micro-object" },
#endif
    POPT_AUTOHELP
    { NULL, 0, 0, NULL, 0 } };

  // Initialise popt library for parsing command-line.
  poptContext optCon( poptGetContext( NULL, argc, (const char **)argv,
                                      optionsTable, 0 ) );
  poptSetOtherOptionHelp( optCon, "[FILE]..." );

  // Initialise libraries Xerces-C and Xalan-C.
  XMLPlatformUtils::Initialize();
  {
    XPathInit oTheXPathInit;

    bool commandLineChecked = false;

    // Catch exceptions from now on.
    try {

      // Parse command-line.
      char c;
      while ( ( c = poptGetNextOpt( optCon ) ) >= 0 );

      // Throw exception on error.
      MMERROR( c >= -1, mimasexception, ,
               poptBadOption( optCon, POPT_BADOPTION_NOALIAS ) << ": "
               << poptStrerror( c ) );

      // Compute framerate.
      double frameRate = (double)frames / SECOND;

      MMERROR( frameDrop == 0 || frames > 0, mimasexception, ,
               "Option '--framedrop' was specified, but frame-rate is "
               "unknown." );

      MMERROR( x11Output == 1 || waitForUser == 0, mimasexception, ,
               "Pause after initialisation only works together with graphical "
               "output." );

      MMERROR( x11Output == 1 || savePics == 0, mimasexception, ,
               "Saving of pictures only works together with graphical "
               "output." );

#ifdef HUMANVISION
      MMERROR( x11Output == 1 || humanVisionTest == "", mimasexception, ,
               "Human vision only supported with X11 output being enabled." );
#endif

      x11_display_ptr display;

      commandLineChecked = true;

      // Load and verify XML-configuration.
      xml_document configuration
        ( string( "http://www.wedesoft.demon.co.uk/2005/micron-vision " ) +
          configSchemaFileName );
      configuration.load( configFileName );

      // Initialise recognition.
#ifdef HUMANVISION
      ex_recognition recognition( configuration.getDocumentElement(),
                                  humanVisionTest, verbose != 0 );
#else
      mi_recognition recognition( configuration.getDocumentElement(),
                                  verbose != 0 );
#endif

      // Get pointer to calibration-object.
      calibration_ptr calib = recognition.get_calibration();

      // Create image-input object.
      shared_ptr< image_input< rgba< unsigned char > > > imageInput =
        open_image_input( selectedInput, optCon,
                          gstreamerSHMKey, gstreamerSEMKey,
                          device, channel,
                          calib->get_width(), calib->get_height(),
                          brightness, contrast, shutter, gain, balance );

      shared_ptr< image_micronoutput< rgba< unsigned char > > > window;

      if ( x11Output ) {

        // Initialise X11-display.

        display = x11_display_ptr( new x11_display );
        window =
          shared_ptr< image_micronoutput< rgba< unsigned char > > >
          ( new image_micronoutput< rgba< unsigned char > >
            ( display, calib->get_width(), calib->get_height() ) );
        cerr << "Warning: Mesa-output may not display properly, "
          "if image is bigger than the screen." << endl;

      };

      // Construct empty scene-object.
      mi_scene_ptr scene( new mi_scene );

      // Counters for acquired and processed images.
      int
        count = 0,
        processed = 0;

      // Keep track of maximum processing time.
      double maxElapsed = 0.0;

      // Display time elapsed.
      if ( verbose )
        cout << "Initialisation took " << ( initialisation.elapsed() )
             << " seconds." << endl;

      // Allocate images.
      image< rgba< unsigned char > > colourImage;

      // Start another timer.
      timer overall;

      // Watch recognition-rates.
      map< string, int > rate;

      // Loop for recognition.
      while ( *imageInput >> colourImage ) {

#ifdef TESTSINGLEFRAME
        for ( int t=0; t<TESTCOUNT + TESTSKIP; t++ ) {
#endif

        image< unsigned char > img( colourImage );

        // Wait at beginning of vision-test.
        if ( waitForUser ) {
          assert( window );
          window->setImage( colourImage );
          display->event_loop();
          waitForUser = 0;
          overall.reset();
        };

        // Skip images in real-time simulation-mode.
        // If frames equals zero, no images will be skipped.
#ifndef TESTSINGLEFRAME
        if ( !frameDrop || overall.elapsed() * SECOND * frameRate <= count ) {
#endif

          // Start another timer.
          timer current;

          double timestamp;
          if ( frames > 0 )
            timestamp = count / frameRate;// Use relative time of capture.
          else
            timestamp = overall.elapsed() * SECOND;// Use relative real time.

          // Generate jobs for all objects.
          // At the moment only one instance per type of micro-object is
          // supported!
          map< string, int > jobs;
          for ( map< string, mi_object_type_ptr >::const_iterator i =
                  recognition.get_object_types().begin();
                i != recognition.get_object_types().end(); i++ )
            jobs[ i->first ] = 1;

          // Remove jobs for objects, which can be tracked.
          for ( vector< mi_object_instance_ptr >::const_iterator i =
                  scene->get_objects().begin();
                i != scene->get_objects().end();
                i++ )
            jobs[ (*i)->get_type() ] = 0;

          // Try recognition of the objects, which have been lost.
          for ( map< string, int >::const_iterator i = jobs.begin();
                i != jobs.end(); i++ )
            if ( i->second > 0 )
              scene->add_job( i->first, 1 );

          // Do the recognition/tracking.
          recognition.recognise( img, timestamp, scene );

          if ( x11Output ) {

            // Setup window for drawing information directly on the screen.
            window->setInfo( scene, timestamp, processed, calib );
            // Display everything.
            assert( window );
            window->setImage( colourImage );

            if ( savePics ) {
              
              assert( x11Output );

              // The information needs to be drawn on the image as well.
              mesa_offscreen< rgba< unsigned char > >
                m( get_pointer( display ), colourImage );
              m.ortho2D();
              displayInfo( m, scene, calib, timestamp, processed,
                           img.getWidth() );
              m.finish();
              
              // Save visualised results for debugging purposes.
              ostringstream outputFileName;
              outputFileName << "debug" << setw( 6 ) << setfill( '0' )
                             << processed << resetiosflags( ios::floatfield )
                             << ".ppm";
              ofstream ofile( outputFileName.str().c_str(), ios::binary );
              MMERROR( ofile << setformat( "PPM" ) << colourImage,
                       mimasexception, ,
                       "Error writing to file \""
                       << outputFileName.str().c_str() << '\"' );
            };

          };

          if ( !x11Output || verbose ) {

            cout << header( scene, timestamp, processed ) << endl;
            
            for ( int i=0; i<min( (signed)scene->get_objects().size(), 100 );
                  i++ ) {

              mi_object_instance_ptr instance( scene->get_objects()[i] );
              cout << instanceInfo( instance ) << endl;
              
            };
            
            cout << endl;
            
          };

          // Increase counter for processed images.
          processed++;

          // Update recognition rates.
          for ( std::vector< mi_object_instance_ptr >::const_iterator
                  i = scene->get_objects().begin();
                i != scene->get_objects().end(); i++ )
            rate[ (*i)->get_type() ]++;

          // Update maximum of processing time.
          double currentElapsed = current.elapsed();
          if ( currentElapsed > maxElapsed )
            maxElapsed = currentElapsed;
        
#ifndef TESTSINGLEFRAME  
        };
#endif

        // Increase counter for acquired images.
        count++;

#ifdef TESTSINGLEFRAME
        if ( count == TESTSKIP ) {
          overall.reset();
          rate.clear();
          processed = 0;
        };
        };
        break;
#endif

      };

      // Display average of time/frame.
      cout << "Average time per frame was "
           << fixed << setw( 6 ) << setprecision( 4 )
           << ( overall.elapsed() / processed ) << " seconds." << endl
           << "Maximum time for a frame was "
           << fixed << setw( 6 ) << setprecision( 4 )
           << maxElapsed << " seconds." << endl;
      if ( processed > 0 ) {
        for ( map< string, int >::const_iterator i = rate.begin();
              i != rate.end(); i++ )
          cout << ( i->second * 100.0 / processed )
               << " % detection-rate for object of type \"" << i->first
               << "\"." << endl;
      };
      cout << resetiosflags( ios::floatfield );

    } catch ( exception &e ) {
      // Display error-message.
      cerr << e.what() << endl;
      // Display command-line information, if error occurred during parsing of
      // command-line.
      if ( !commandLineChecked )
        poptPrintUsage( optCon, stderr, 0 );
      // Indicate error in return value.
      retVal = 1;
    };

  };
  // Call destructors of Xerces-C and Xalan-C library.
  XMLPlatformUtils::Terminate();
 
  // Call destructor of command-line parser.
  poptFreeContext( optCon );

  return retVal;
}
