#include <cassert>
#include <mimas/xml_node_reference_list.h>
#include "bounded_hough_plugin.hh"
#include "tools.hh"

using namespace boost::numeric;
using namespace mimas;
using namespace std;

#define VISIONTYPE "bounded-hough"

bounded_hough_plugin::bounded_hough_plugin
( const calibration_ptr &_calib,
  const xml_reference< XalanElement > &configuration,
  bool _verbose ):
  recognition_method_4dof( _calib, Tracking, configuration, _verbose )
{
  // Read global parameters.
  xml_reference< XalanElement >
    global_config( dynamic_pointer_cast< XalanElement >
                   ( configuration.selectNode( "methods/"VISIONTYPE ) ) );

  fringe = atoi( global_config.selectNode( "fringe" ).
                 getNodeText().c_str() );
  sampling_rate = atoi( global_config.selectNode( "sampling_rate" ).
                        getNodeText().c_str() );
  min_gradient = atoi( global_config.selectNode( "min_gradient" ).
                       getNodeText().c_str() );

  xml_node_reference_list objects = configuration.selectNodes
    ("database/object-type/tracking/"VISIONTYPE );

  for ( int i=0; i<objects.getLength(); i++ ) {

    xml_reference< XalanElement > item
      ( dynamic_pointer_cast< XalanElement >( objects.item( i ) ) );

    string className( item.selectNode( "../../@name" ).getNodeText() );

    assert( models.find( className ) == models.end() );
    models[ className ] =
      hough_table_ptr
      ( new hough_table( calib, item, sampling_rate, fringe, min_gradient,
                         verbose ) );

  };
}

vector< match > bounded_hough_plugin::recognise
( const image< unsigned char > &img, double timestamp,
  mi_scene_ptr scene ) const
{
  if ( verbose ) cout << "   "VISIONTYPE":" << endl;
  vector< match > retVal;

  image< bool > scene_features
    ( edgeDetect( down_sample( img, sampling_rate, fringe ),
                  min_gradient ) );

  ublas::vector< int > offset( 2 );
  offset(0) = ( calib->get_middle_x() - fringe ) / sampling_rate;
  offset(1) = ( calib->get_middle_y() - fringe ) / sampling_rate;

  vector< ublas::vector< int > >
    points( scalePoints( subtractOffset( featurePoints( scene_features ),
                                         offset ),
                         1.0 / calib->get_scale() ) );

  if ( !points.empty() ) {

    for ( vector< mi_object_instance_ptr >::const_iterator i =
            scene->get_objects().begin(); i != scene->get_objects().end();
          i++ ) {

      std::string type = (*i)->get_type();
      std::map< std::string, hough_table_ptr >::const_iterator model =
        models.find( type );

      if ( model != models.end() ) {
      
        if ( verbose )
          cout << "      object of type \"" << type << "\":" << endl;

        pair< mi_pose, double > result
          ( model->second->
            recognise( points,
                       convert_pose_to_internal( type, (*i)->get_pose() ) ) );
        
        if ( verbose ) {
          cout << "         percentage of matched features                                  :" << result.second;
          if ( result.second >= model->second->get_threshold() )
            cout << " (accepted)" << endl;
          else
            cout << " (rejected)" << endl;
        };

        if ( result.second >= model->second->get_threshold() )
          retVal.push_back
            ( match( type, convert_pose_to_external( type, result.first ),
                     result.second ) );
        
      };
      
    };
  } else
    if ( verbose )
      cout << "      Image doesn't have any features." << endl;

  return retVal;
}
