#include <mimas/xml_node_reference_list.h>
#include "hough_table.hh"
#include "tools.hh"
#include "units.h"

using namespace boost;
using namespace boost::numeric;
using namespace mimas;
using namespace std;

hough_table::hough_table( const calibration_ptr &_calib,
                          const mimas::xml_reference< XalanElement >
                          &boundedHoughNode, int _samplingRate, int _fringe,
                          short int minGradient, bool _verbose ):
  calib(_calib), sampling_rate(_samplingRate), fringe(_fringe),
  max_num_features( 0 ), verbose(_verbose)
{
  skip = atoi( boundedHoughNode.selectNode( "stack/skip" ).
               getNodeText().c_str() );
  int count = atoi( boundedHoughNode.selectNode( "stack/count" ).
                    getNodeText().c_str() );
  z_resolution = atof( boundedHoughNode.selectNode( "../../shared/stack/z-resolution" ).
                       getNodeText().c_str() ) * MICROMETER;
  threshold = atof( boundedHoughNode.selectNode( "threshold" ).
                    getNodeText().c_str() );
  int xyRange2 = atoi( boundedHoughNode.selectNode( "range2/xy" ).
                       getNodeText().c_str() );
  int zRange2 = atoi( boundedHoughNode.selectNode( "range2/z" ).
                      getNodeText().c_str() );
  double angleRange2 = atof( boundedHoughNode.selectNode( "range2/angle" ).
                             getNodeText().c_str() ) * M_PI / 180.0;
  int antialias = atoi( boundedHoughNode.selectNode( "hough/antialias" ).
                       getNodeText().c_str() );
  int anglesub = atoi( boundedHoughNode.selectNode( "hough/anglesub" ).
                       getNodeText().c_str() );
  xml_node_reference_list images
    ( boundedHoughNode.selectNodes( "../../shared/stack/image" ) );

  MMERROR( skip < images.getLength(), mimasexception, ,
           "Can not skip " << skip << " images, because stack only is "
           "containing " <<images.getLength() << " images." );
  MMERROR( skip + count <= images.getLength(), mimasexception, ,
           "Can not take " << ( skip + 1 ) << "th to " << ( skip + count )
           << "th image of stack with size " << images.getLength() );
  MMERROR( 2 * xyRange2 + 1 >= anglesub, mimasexception, ,
           "Cannot estimate sub-pixel shifts. Current value of "
           "'hough/anglesub' (" << anglesub << ") is to high for this value "
           "of 'range2/xy' (" << xyRange2 << ")" );

  ublas::vector< double > centre( 2 );

  vector< vector< ublas::vector< int > > > points;
  for ( int i=skip; i<skip+count; i++ ) {

    assert( i >= 0 && i < images.getLength() );
    string fileName( images.item( i ).getNodeText() );

    image< unsigned char > img
      ( down_sample( load_image< unsigned char >( fileName ),
                     sampling_rate, 0 ) );

    image< bool > features( edgeDetect( img, minGradient ) );
    vector< ublas::vector< int > > currentPoints( featurePoints( features ) );

    // Keep track of maximum number.
    if ( (signed)currentPoints.size() > max_num_features )
      max_num_features = currentPoints.size();

    points.push_back( currentPoints );

    if ( i == skip ) {
      width = img.getWidth();
      height = img.getHeight();
    };
  };

  hough_data.resize( count );
  for ( int i=0; i<count; i++ )
    hough_data[i].resize( extents[ height ][ width ] );
  
  centre[0] = width  / 2.0;
  centre[1] = height / 2.0;

  double angleStep;
  if ( anglesub > 1 )
    angleStep = 2.0 * angleRange2 / ( anglesub - 1 );
  else
    angleStep = 0.0;

  mi_pose
    center( mi_pose::translate_3d
            ( centre[0], centre[1], 0.0 ) ),
    scale( mi_pose::scale_3d
           ( calib->get_pixel_size() * sampling_rate,
             calib->get_pixel_size() * sampling_rate,
             -z_resolution ) );

  for ( int dangleI=0; dangleI<anglesub; dangleI++ ) {
    double dangle;
    if ( anglesub > 1 )
      dangle = -angleRange2 + dangleI * angleStep;
    else
      dangle = 0.0;
    mi_pose stirR( mi_pose::rotate_3d( dangle, 0.0, 0.0 ) );
    for ( int dy=-xyRange2; dy<=xyRange2; dy++ ) {
      for ( int dx=-xyRange2; dx<xyRange2; dx++ ) {
        mi_pose stirXYR( mi_pose::translate_3d( dx, dy, 0 ) * stirR );
        for ( int dz=-zRange2; dz<=zRange2; dz++ ) {
          mi_pose stirZXYR( mi_pose::translate_3d( 0, 0, dz ) * stirXYR );
          mi_pose stir( center *  stirZXYR * center.inv() );
          int index = change.size();
          for ( int i=max( 0, -dz ); i<min( count, count - dz ); i++ ) {
            for ( vector< ublas::vector< int > >::const_iterator j =
                    points[i+dz].begin(); j != points[i+dz].end(); j++ ) {
              float weight;
              double radius = norm_2( *j - centre );
              if ( anglesub > 1 )
                weight = angleStep * radius;
              else
                weight = 1.0;
              if ( anglesub == 1 || weight <= 1.0 ) {
              // if ( anglesub == 1 ) {
                ublas::vector< int > mapped( stir.pose2d().map( *j ) );
                if ( mapped( 0 ) >= 0 && mapped( 0 ) < width &&
                     mapped( 1 ) >= 0 && mapped( 1 ) < height ) {
                  ( hough_data[i][ mapped( 1 ) ][ mapped( 0 ) ] )[ change.size() ] +=
                    weight;
                };
              } else {
                // } else if ( weight >= 1.0 ) {
                int ext = (int)( weight / 2.0 );
                for ( int k=-ext; k<=ext; k++ ) {
                  mi_pose stirX( stir * mi_pose::rotate_3d( k / radius, 0.0, 0.0 ) );
                  ublas::vector< int > mapped( stirX.pose2d().map( *j ) );
                  if ( mapped( 0 ) >= 0 && mapped( 0 ) < width &&
                       mapped( 1 ) >= 0 && mapped( 1 ) < height ) {
                    ( hough_data[i][ mapped( 1 ) ][ mapped( 0 ) ] )[ change.size() ] +=
                      1.0;
                  };
                };
              };
            };
          };
          change.push_back( scale * stir * scale.inv() );
        };
      };
    };
  };
}

pair< mi_pose, double > hough_table::recognise
( const vector< ublas::vector< int > > &points,
  const mi_pose &pose )
{
  pair< mi_pose, double > retVal;

  int depth = (int)round( -pose.getTranslation()[2] / z_resolution ) - skip;
  if ( depth >= 0 && depth < hough_data.size() ) {
    mi_pose scale( mi_pose::scale_3d( sampling_rate * calib->get_pixel_size(),
                                      sampling_rate * calib->get_pixel_size(),
                                      1.0 ) );
    mi_pose pixelMap( ( scale.inv() * pose.inv() * scale ).pose2d() );
    vector< float > hough_space( change.size() );
    for ( vector< ublas::vector< int > >::const_iterator j =
            points.begin(); j != points.end(); j++ ) {
      
      ublas::vector< int > mapped( pixelMap.map( *j ) );
      if ( mapped( 0 ) >= 0 && mapped( 0 ) < width &&
           mapped( 1 ) >= 0 && mapped( 1 ) < height ) {
        
        const map< int, float > *votes =
          &hough_data[ depth ][ mapped( 1 ) ][ mapped( 0 ) ];
        
        for ( map< int, float >::const_iterator k = votes->begin();
              k != votes->end(); k++ )
          hough_space[ k->first ] += k->second;
        
      };
    };
    
    vector< float >::const_iterator m = max_element
      ( hough_space.begin(), hough_space.end() );
    
    if ( m != hough_space.end() ) {
      
      int index = m - hough_space.begin();
      retVal.first = pose * change[ index ];
      retVal.second = (double)(*m) / max_num_features;
      
    } else
      
      retVal.second = -1.0;

  } else
    retVal.second = -1.0;

  return retVal;
}
