// #define DESPECKLE

#ifndef NDEBUG
#include <iostream>
#endif
#include <iomanip>
#ifdef DESPECKLE
#include <mimas/mm_image_funcs.h>
#endif
#include <mimas/timer.h>
#include <mimas/xml_node_reference_list.h>
#include <sstream>
#include "geometric_map_xyza.hh"
#include "geometric_map_xyz.hh"
#include "geometric_voting_plugin.hh"
#include "units.h"

using namespace boost;
using namespace boost::numeric;
using namespace mimas;
using namespace std;

// ./voting -a 2000 -p 7 -i 0.1 -t 23 -l 30 -u 34 -m sphere/shift/model.ppm sphere/shift/snap-unknown-20041029-1234*

#define VISIONTYPE "geometric-voting"

geometric_voting_plugin::geometric_voting_plugin
( const calibration_ptr &_calib,
  const xml_reference< XalanElement > &configuration,
  bool _verbose ):
  recognition_method_4dof( _calib, Recognition, configuration, _verbose )
#ifndef NDEBUG
  , display( &dpy )
#endif
{
  // Read global parameters.
  xml_reference< XalanElement >
    global_config( dynamic_pointer_cast< XalanElement >
                   ( configuration.selectNode( "methods/"VISIONTYPE ) ) );
  fringe = atoi( global_config.selectNode( "fringe" ).
                 getNodeText().c_str() );
  sampling_rate = atoi( global_config.selectNode( "sampling_rate" ).
                        getNodeText().c_str() );
  timeout = atof( global_config.selectNode( "timeout" ).
                  getNodeText().c_str() );
  min_gradient = atoi( global_config.selectNode( "min_gradient" ).
                       getNodeText().c_str() );

  xml_node_reference_list objects
    ( configuration.selectNodes
      ( "database/object-type/recognition/"VISIONTYPE ) );

  std::map< std::string, geometric_map_ptr > hash_map;

  for ( int i=0; i<objects.getLength(); i++ ) {

    xml_reference< XalanElement > item
      ( dynamic_pointer_cast< XalanElement >( objects.item( i ) ) );

    string
      className( item.selectNode( "../../@name" ).getNodeText() ),
      hash_type( item.selectNode( "@dof" ).getNodeText() );

    if ( !hash_map[ hash_type ] ) {

      MMERROR( global_config.existNode( hash_type.c_str() ),
               mimasexception, ,
               "Global configuration-block \"methods/geometric-voting/"
               << hash_type << "\" is missing." );

      xml_reference< XalanElement >
        hash_config( dynamic_pointer_cast< XalanElement >
                     ( global_config.selectNode( hash_type.c_str() ) ) );

      if ( hash_type == "xyza" )
        hash_map[ hash_type ] =
          geometric_map_ptr( new geometric_map_xyza( hash_config, verbose ) );
      else if ( hash_type == "xyz" ) {
        hash_map[ hash_type ] =
          geometric_map_ptr( new geometric_map_xyz( hash_config, verbose ) );
      } else {
        assert( false );
      };
    };

    assert( models.find( className ) == models.end() );
    models[ className ] =
      voting_table_ptr
      ( new voting_table( calib, item, sampling_rate, fringe, min_gradient,
                          hash_map[ hash_type ], verbose ) );

  };
}

vector< match > geometric_voting_plugin::recognise
( const image< unsigned char > &img, double timestamp,
  mi_scene_ptr scene ) const
{
  if ( verbose ) cout << "   "VISIONTYPE":" << endl;

  vector< match > retVal;

  if ( !scene->get_jobs().empty() ) {

#ifndef DESPECKLE
    image< bool > scene_features
      ( edgeDetect( down_sample( img, sampling_rate, fringe ),
                    min_gradient ) );
#else
    image< unsigned char > processed;
    despeckleMedian( down_sample( img, sampling_rate, fringe ), processed );
    image< bool > scene_features( edgeDetect( processed, min_gradient ) );
#endif

#ifndef NDEBUG
    display << tobw( scene_features );

    cerr << "Assumed scaling factor of image is " << calib->get_scale()
         << endl;
#endif
    
    ublas::vector< int > offset( 2 );
    offset(0) = ( calib->get_middle_x() - fringe ) / sampling_rate;
    offset(1) = ( calib->get_middle_y() - fringe ) / sampling_rate;

    vector< ublas::vector< int > >
      points( scalePoints( subtractOffset( featurePoints( scene_features ),
                                           offset ),
                           1.0 / calib->get_scale() ) );
    
#ifndef NDEBUG
    cerr << points.size() << " scene features" << endl;
#endif

    if ( !points.empty() ) {
      
      // Create map of statistics.
      map< string, geometric_voting_statistics > statistics;

      // Make set of micro-object types to be searched in the scene-image.
      map< string, voting_table_ptr > activeModels;
      for ( map< string, int >::const_iterator i = scene->get_jobs().begin();
            i != scene->get_jobs().end(); i++ )
        if ( models.find( i->first ) != models.end() )
          activeModels[ i->first ] = models.find( i->first )->second;

      timer t;

      do {

        // Generate set of random transforms for geometric hashing.
        map< string, mi_pose_ptr > pose;

        // Process this set.
        for ( map< string, voting_table_ptr >::const_iterator i =
                activeModels.begin(); i != activeModels.end(); i++ ) {
          
          voting_table_ptr hash_table( i->second );
          geometric_map_ptr hash_map( hash_table->get_hash_map() );
          string hash_type = hash_map->get_hash_type();
          
          if ( !pose[ hash_type ] ) {
            pose[ hash_type ] = mi_pose_ptr( new mi_pose );
            if ( !hash_map->random_poses( points, *pose[ hash_type ] ) )
              pose[ hash_type ].reset();
          };

          mi_pose_ptr currentPose( pose[ hash_type ] );
          
          if ( currentPose ) {
            // Apply all random-transforms.
            vector< pair< mi_pose, double > > result
              ( i->second->recognise( scene_features, points,
                                      *currentPose,
                                      &statistics[ i->first ] ) );
            // Convert each resulting pose to a match-object.
            for ( int k=0; k<(signed)result.size(); k++ )
              retVal.push_back
                ( match( i->first,
                         convert_pose_to_external( i->first,
                                                   result[k].first ),
                         result[k].second ) );
          };

        }

      } while ( t.elapsed() < timeout );

      if ( verbose ) {
        for ( map< string, voting_table_ptr >::const_iterator i =
                activeModels.begin(); i != activeModels.end(); i++ ) {
          const geometric_voting_statistics *statistic =
            &statistics[ i->first ];
          cout << "      object \"" << i->first << "\":" << endl
               << "         min_hits reached (<=max_tupples)"
                  "                                : "
               << statistic->minHitsCounter << " time(s) (maximum hits was "
               << statistic->maxHitsReached << ")" << endl
               << "         max_misses ok as well"
                  "                                           : "
               << statistic->maxMissesCounter
               << " time(s) (minimum misses was "
               << statistic->minMissesReached << ")" << endl;
        }
      };

    } else
      if ( verbose )
        cout << "      Image doesn't have any features." << endl;
  };

  return retVal;
}

