#ifndef __GEOMETRIC_VOTING_H
#define __GEOMETRIC_VOTING_H

#include <map>
#include <string>
#ifndef NDEBUG
#include <mimas/image_mesaoutput.h>
#endif
#include <mimas/xml_reference.h>
#include "geometric_map.hh"
#include "recognition_method_4dof.hh"
#include "tools.hh"
#include "voting_table.hh"

/** Recognition-method geometric-voting.
    See chapter about <B>geometric hashing</B> in the computer vision book
    from Prentice Hall:
    \verbatim
    Computer Vision - A Modern Approach
    Prentice Hall 2002
    \endverbatim
    @author Jan Wedekind (j.wedekind@shu.ac.uk) */
class geometric_voting_plugin: public recognition_method_4dof
{
public:
  /** Constructor.
      The constructor will extract the configuration from the XML-tree.
      For all micro-objects to be recognized with this method, voting_table
      objects are created.
      @param _calib Calibration data.
      @param configuration Top-element of vision-configuration.
      @param _verbose Set to \c true, if you want to get verbose information
      from the vision-algorithms. */
  geometric_voting_plugin( const calibration_ptr &_calib,
                           const mimas::xml_reference< XalanElement >
                           &configuration,
                           bool _verbose );
  /** Recognise objects using geometric voting.
      @param image Image to recognise objects in.
      @param timestamp Timestamp for image
      @param scene Scene information. */
  virtual std::vector< match > recognise
  ( const mimas::image< unsigned char > &image,
    double timestamp,
    mi_scene_ptr scene ) const;
protected:
  /** Width of boundary to cut from image.
      Cutting of the boundary is required, when the image-source generates a
      black frame, which introduces strong gradient-features. */
  int fringe;
  /** Sampling rate for sub-sampling image.
      For performance reasons the image can be sub-sampled before processing
      with the specified sampling-rate. */
  int sampling_rate;
  /// Timeout for recognition.
  double timeout;
  /// Gradient threshold for feature extraction.
  short int min_gradient;
  /** Map of voting tables.
      The name of a micro-object-type is mapped on a voting_table.
      @see voting_table */
  std::map< std::string, voting_table_ptr > models;
#ifndef NDEBUG
  ///
  mutable mimas::x11_display dpy;
  /// Display for debugging.
  mutable mimas::image_mesaoutput< mimas::rgba< unsigned char > > display;
#endif
};

#endif
