#include <boost/numeric/ublas/banded.hpp>
#include <cassert>
#include <cmath>
#include <mimas/xml_node_reference_list.h>
#include "geometric_voting_statistics.hh"
#include "tools.hh"
#include "units.h"
#include "voting_table.hh"

using namespace boost;
using namespace boost::numeric;
using namespace mimas;
using namespace std;


voting_table::voting_table( const calibration_ptr &_calib,
                            const mimas::xml_reference< XalanElement >
                            &geometricVotingNode, int _sampling_rate,
                            int _fringe, short int minGradient,
                            geometric_map_ptr _hash_map,
                            bool _verbose ):
  calib( _calib ), sampling_rate( _sampling_rate ), fringe( _fringe ),
  max_num_features( 0 ), hash_map( _hash_map ), verbose( _verbose )
#ifndef NDEBUG
  , display( &dpy )
#endif
{
  min_hits = atof( geometricVotingNode.selectNode( "min_hits" ).getNodeText().
                   c_str() );
  max_misses = atof( geometricVotingNode.selectNode( "max_misses" ).
                    getNodeText().c_str() );
  skip = atoi( geometricVotingNode.selectNode( "stack/skip" ).
               getNodeText().c_str() );
  int count = atoi( geometricVotingNode.selectNode( "stack/count" ).
                    getNodeText().c_str() );
  z_resolution = atof( geometricVotingNode.selectNode( "../../shared/stack/z-resolution" ).
                      getNodeText().c_str() ) * MICROMETER;
  xml_node_reference_list images
    ( geometricVotingNode.selectNodes( "../../shared/stack/image" ) );

  MMERROR( skip < images.getLength(), mimasexception, ,
           "Can not skip " << skip << " images, because stack only is "
           "containing " <<images.getLength() << " images." );
  MMERROR( skip + count <= images.getLength(), mimasexception, ,
           "Can not take " << ( skip + 1 ) << "th to " << ( skip + count )
           << "th image of stack with size " << images.getLength() );

  int mw = -1, mh = -1, mx = 0, my = 0;

  for ( int i=skip; i<skip+count; i++ ) {

    assert( i >= 0 && i < images.getLength() );

    string fileName( images.item( i ).getNodeText() );
    image< unsigned char > img
      ( down_sample( load_image< unsigned char >( fileName ),
                     sampling_rate, 0 ) );

    if ( i == skip ) {
      // Initialise hash-table.
      mw = img.getWidth();
      mh = img.getHeight();
      mx = mw / 2;
      my = mh / 2;
      hash_table.resize( extents[ mh ][ mw ] );
    } else {
      MMERROR( img.getWidth() == mw && img.getHeight() == mh,
               mimasexception, ,
               ( skip + 1 ) << "th image of stack was of size "
               << mw << "x" << mh << ", but " << ( i + 1 ) << "th image is "
               "of size " << img.getWidth() << "x" << img.getHeight()
               << "." );
    };

#ifndef NDEBUG
    cerr << "minGradient is " << minGradient << endl;
#endif
    image< bool > features( edgeDetect( img, minGradient ) );
  
#ifndef NDEBUG
    display << tobw( features );
#endif

    points_stack.push_back( featurePoints( features ) );

    const vector< ublas::vector< int > > *points = &points_stack.back();

    // Keep track of maximum number.
    if ( (signed)points->size() > max_num_features )
      max_num_features = points->size();

    vector< mi_pose > currentCandidates( hash_map->candidates( *points ) );

    for ( int j=0; j<(signed)currentCandidates.size(); j++ ) {
      
      const mi_pose *pose = &currentCandidates[ j ];

      for ( int k=0; k<(signed)points->size(); k++ ) {

        assert( mw > 0 && mh > 0 );

        ublas::vector< int > x( pose->map( (*points)[k] ) );

        if ( x[0] + mx >= 0 && x[0] + mx < mw &&
             x[1] + my >= 0 && x[1] + my < mh )
          ( hash_table[ x[1] + my ][ x[0] + mx ] )[ candidates.size() ]++;

      };

      candidates.push_back( pose3d( *pose, 0, 0, i ) );

    };
    
  };
  
}

vector< pair< mi_pose, double > > voting_table::recognise
( const image< bool > &scene_features,
  const vector< ublas::vector< int > > &points,
  const mi_pose &poseInv, geometric_voting_statistics *statistic ) const
{
  vector< int > accumulator( candidates.size() );

  const int
    mw = hash_table.shape()[1],
    mh = hash_table.shape()[0],
    mx = mw / 2,
    my = mh / 2;

  for ( int k=0; k<(signed)points.size(); k++ ) {
    ublas::vector< int > x( poseInv.map( points[k] ) );
    if ( x[0] + mx >= 0 && x[0] + mx < mw &&
         x[1] + my >= 0 && x[1] + my < mh ) {
      const map< int, int > *entry = &hash_table[ x[1] + my ][ x[0] + mx ];
      for ( map< int, int >::const_iterator i = entry->begin();
            i != entry->end(); i++ )
        accumulator[ i->first ] += i->second;
    };
  };

  vector< pair< mi_pose, double > > retVal;

#ifndef NDEBUG
  cerr << "Maximum number of features: " << max_num_features << endl;
#endif

  for ( int i=0; i<(signed)accumulator.size(); i++ ) {

    if ( accumulator[i] >= statistic->maxHitsReached * max_num_features &&
         max_num_features > 0 )
      statistic->maxHitsReached = (double)accumulator[i] / max_num_features;
    
    if ( accumulator[i] >= min_hits * max_num_features ) {

      statistic->minHitsCounter++;
      mi_pose objectPose
        ( pose3d( poseInv.inv(),
                  0, 0, 0 ) *
          candidates[i] );
#ifndef NDEBUG
      cerr << accumulator[i] << '/' << max_num_features << " hits" << endl;
#endif
      if ( verification( scene_features, objectPose, statistic ) )
        retVal.push_back( make_pair( objectPose, accumulator[i] ) );
      
    };
  };

  return retVal;
}

mi_pose voting_table::pose2d( const mi_pose &pose ) const
{
  assert( pose.getRotation().size1() == 3 );
  assert( pose.getTranslation().size() == 3 );

  ublas::matrix< double > rotation2d( 2, 2 );
  rotation2d( 0, 0 ) = pose.getRotation()( 0, 0 ) * calib->get_scale();
  rotation2d( 0, 1 ) = pose.getRotation()( 0, 1 ) * calib->get_scale();
  rotation2d( 1, 0 ) = pose.getRotation()( 1, 0 ) * calib->get_scale();
  rotation2d( 1, 1 ) = pose.getRotation()( 1, 1 ) * calib->get_scale();

  ublas::vector< double > translation2d( 2 );
  translation2d( 0 ) =
    ( pose.getTranslation()( 0 ) / calib->get_pixel_size() +
      calib->get_middle_x() - fringe ) / sampling_rate;
  translation2d( 1 ) =
    ( pose.getTranslation()( 1 ) / calib->get_pixel_size() +
      calib->get_middle_y() - fringe ) / sampling_rate;

  return mi_pose( rotation2d, translation2d );
}

int voting_table::pose_depth( const mi_pose &pose ) const
{
  assert( pose.getRotation().size1() == 3 );
  assert( pose.getTranslation().size() == 3 );
  int retVal = (int)round( -pose.getTranslation()[2] / z_resolution ) - skip;
  assert( retVal >= 0 );
  assert( retVal < (signed)points_stack.size() );
  return retVal;
}

mi_pose voting_table::pose3d( const mi_pose &pose, int x, int y, int z ) const
{
  assert( pose.getRotation().size1() == 2 );
  // It's better to check twice.
  assert( pose.getRotation().size1() == 2 );
  assert( pose.getTranslation().size() == 2 );

  ublas::matrix< double > rotation3d( 3, 3 );
  rotation3d( 0, 0 ) = pose.getRotation()( 0, 0 );
  rotation3d( 0, 1 ) = pose.getRotation()( 0, 1 );
  rotation3d( 0, 2 ) = 0.0;
  rotation3d( 1, 0 ) = pose.getRotation()( 1, 0 );
  rotation3d( 1, 1 ) = pose.getRotation()( 1, 1 );
  rotation3d( 1, 2 ) = 0.0;
  rotation3d( 2, 0 ) = 0.0;
  rotation3d( 2, 1 ) = 0.0;
  rotation3d( 2, 2 ) = 1.0;

  ublas::vector< double > translation3d( 3 );
  translation3d( 0 ) =
    ( pose.getTranslation()( 0 ) * sampling_rate - x ) *
    calib->get_pixel_size() * calib->get_scale();
  translation3d( 1 ) =
    ( pose.getTranslation()( 1 ) * sampling_rate - y ) *
    calib->get_pixel_size() * calib->get_scale();
  translation3d( 2 ) = - z * z_resolution;

  return mi_pose( rotation3d, translation3d );
}

bool voting_table::verification( const image< bool > &scene_features,
                                 const mi_pose &objectPose,
                                 geometric_voting_statistics *statistic ) const
{
  mi_pose pose( pose2d( objectPose ) );
  int z = pose_depth( objectPose );

#ifndef NDEBUG
  image< bool > mapped( scene_features );
#endif

  const vector< ublas::vector< int > > *slice = &points_stack[ z ];

  int nMisses = 0;

  for ( vector< ublas::vector< int > >::const_iterator i=slice->begin();
        i != slice->end(); i++ ) {
    ublas::vector< int > p( pose.map( *i ) );
    if ( p[0] >= 0 && p[1] >= 0 &&
         p[0] < (signed)scene_features.getWidth() &&
         p[1] < (signed)scene_features.getHeight() ) {
#ifndef NDEBUG
      mapped.pixel( p[0], p[1] ) = true;
#endif
      if ( !scene_features.pixel( p[0], p[1] ) ) nMisses++;
    } else
      nMisses++;
  };

#ifndef NDEBUG
  display << image< rgba< unsigned char > >( tobw( mapped ) );
  cerr << ' ' << nMisses << '/' << max_num_features << " misses" << endl;
#endif

  bool retVal;
  if ( nMisses <= statistic->minMissesReached * max_num_features &&
       max_num_features > 0 )
    statistic->minMissesReached = (double)nMisses / max_num_features;

  if ( nMisses <= max_misses * max_num_features ) {
    statistic->maxMissesCounter++;
    retVal = true;
  } else
    retVal = false;

  return retVal;
}
