#ifndef __VOTING_TABLE_H
#define __VOTING_TABLE_H

#include <boost/smart_ptr.hpp>
#include <boost/multi_array.hpp>
#include <boost/numeric/ublas/vector.hpp>
#include <map>
#include <mimas/image.h>
#ifndef NDEBUG
#include <mimas/image_mesaoutput.h>
#endif
#include <mimas/object.h>
#include <mimas/xml_reference.h>
#include <utility>
#include <vector>
#include "calibration.hh"
#include "geometric_map.hh"
#include "geometric_voting_statistics.hh"
#include "mi_pose.hh"

/** Array being used in geometric hashing.
    An object of this class holds the table for recognizing one object with
    four degrees of freedom.
    @see geometric_voting
    @author Jan Wedekind (j.wedekind@shu.ac.uk) */
class voting_table: public mimas::object
{
 public:
  /** Configure table.
      Creates the table with votes for a single type of micro-object.
      @param geometricVotingNode XML-node referencing a node with X-path
      <B>database/object-type/recognition/geometric-voting</B>.
      @param _calib Calibration data.
      @param geometricVotingNode Root-node of part of XML-document for
      initialisation.
      @param _samplingRate Sampling rate for sub-sampling the images.
      @param _fringe Size of fringe to be discarded from image.
      @param minGradient Threshold to select gradient-points.
      @param _hash_map Object for computing hash-functions.
      @param _verbose Set to \c true, if you want to get verbose information
      from the vision-algorithms. */
  voting_table( const calibration_ptr &_calib,
                const mimas::xml_reference< XalanElement >
                &geometricVotingNode, int _samplingRate, int _fringe,
                short int minGradient, geometric_map_ptr _hash_map,
                bool _verbose );
  /** Generate match for a point-pair.
      This method finds the best match for a single point-pair. The number
      of votes for all possible poses is accumulated and the best is returned.
      @param scene_features Image of features.
      @param points List of the 2-D features.
      @param poseInv randomly generated pose for using as hash-function.
      @param maxHitsReached This is used for generating verbose output only.
      @param minHitsCounter This is used for generating verbose output only.
      @param minMissesReached This is used for generating verbose output only.
      @param maxMissesCounter This is used for generating verbose output only.
      @return Vector containing poses and fit-values of the matches with
      sufficient hits.
      @see min_hits */
  std::vector< std::pair< mi_pose, double > > recognise
    ( const mimas::image< bool > &scene_features,
      const std::vector< boost::numeric::ublas::vector< int > > &points,
      const mi_pose &poseInv,
      geometric_voting_statistics *statistic ) const;
  /** Z-Resolution.
      Get distance between the focussed planes of two consecutive images of
      the focus stack.
      @return z-resolution of focus stack. */
  double get_z_resolution(void) const { return z_resolution; }
  /** Get pointer to object for generating hash-functions.
      @see geometric_map */
  geometric_map_ptr get_hash_map(void) const { return hash_map; }
 protected:
  /** Convert 4-DOF transformation to two-dimensional transformation.
      This does the inverse of the \c pose3d method.
      @param pose 3-dimensional transformation.
      @see pose3d */
  mi_pose pose2d( const mi_pose &pose ) const;
  /** Extract stack-level from 4-DOF transformation.
      The following condition always should hold:
      \code
      pose_depth( pose3d( anyPose, z ) ) == z
      \endcode */
  int pose_depth( const mi_pose &pose ) const;
  /** Convert two-dimensional transformation to transformation with 4-DOF.
      The transformation-matrix and -vector are made 3-dimensional. The
      4th degree of freedom (z-translation) is added.
      @param pose 2-dimensional transformation.
      @param x x-offset (pixels) to be subtracted from coordinates.
      @param y y-offset (pixels) to be subtracted from coordinates.
      @param z Additional unit-less translation in z-direction.
      @see units
      @see pose2d */
  mi_pose pose3d( const mi_pose &pose, int x, int y, int z ) const;
  /** Verify a hypothesis.
      After a hypothesis was formed, it is verified by projecting the
      object's features on the image and checking, wether most of them are
      there.
      @param scene_features Image of scene-features.
      @param objectPose 4-DOF pose of object.
      @param minMissesReached This is used for generating verbose output only.
      @param maxMissesCounter This is used for generating verbose output only.
  */
  bool verification( const mimas::image< bool > &scene_features,
                     const mi_pose &objectPose,
                     geometric_voting_statistics *statistic ) const;
  /// Calibration data.
  calibration_ptr calib;
  /// Sampling-rate.
  int sampling_rate;
  /// Fringe.
  int fringe;
  /** Minimum fraction of matching features for a hypothesis.
      A match with a lower number of hits has to be discarded immediately. */
  double min_hits;
  /** Maximum fraction of unmatched features for a hypothesis.
      The maximum allowed number of features not being found in the scene.
      This number is used in the verification step.
      @see verification */
  double max_misses;
  /** Number of images being skipped from shared focus stack.
      An additional offset has to be taken into account, if only a part of the
      original focus stack is being used. */
  int skip;
  /** Z-Resolution of focus stack.
      Distance between the focussed planes of two consecutive images of the
      focus stack. */
  double z_resolution;
  /** Feature lists.
      The features of each image of the micro-object's focus-stack is stored
      as an element of this list. */
  std::vector< std::vector< boost::numeric::ublas::vector< int > > >
    points_stack;
  /** Maximum number of features.
      Number of features of the stack-image, with has the most. */
  int max_num_features;
  /** Table with votes.
      The table is two-dimensional and each element is a map, assigning
      mapping candidate-indices on number of votes.
      @see candidates */
  boost::multi_array< std::map< int, int >, 2 > hash_table;
  /** Candidate list.
      Each candidate is a different possible pose of the object (relative to
      a randomly selected point of pairs). */
  std::vector< mi_pose > candidates;
  /** Hash functions.
      Functionality for generating transformations from features. */
  geometric_map_ptr hash_map;
  ///
  bool verbose;
#ifndef NDEBUG
  ///
  mutable mimas::x11_display dpy;
  /// Display for debugging purposes.
  mutable mimas::image_mesaoutput< mimas::rgba< unsigned char > > display;
#endif
};

/// Pointer to voting_table object.
typedef boost::shared_ptr< voting_table > voting_table_ptr;

#endif
