/**
 * Stencil tracking with 4 DoF (2 translation, a rotation and a zoom.)
 *
 *
 *
 * Time-stamp: <05/10/19 21:06:11 engmb>
 * Copyright (C) 2005 by Manuel Boissenin
 *
 * Manuel.Boissenin@gmail.com
 * @author Manuel Boissenin
 * @date 17/01/2005
 *
 ***********************************************************************/
#include <QtCore/QTime>
#ifndef NDEBUG
#include <boost/multi_array.hpp>
#endif
#include <utility>
#include "hough_tracking.hh"

#ifndef NDEBUG
using namespace boost;
#endif

//add a mechanism that will discard an image model if its number of feature is less
//than a threshold value.

hough_tracking::hough_tracking( const std::vector< std::vector<Feature> > &feature_images,
				int model_width, int model_height, int K ,
				int x_min, int x_max, int x_sub,
				int y_min, int y_max, int y_sub,
				float theta_min, float theta_max, int theta_sub,
				int nb_thread):
  model_width(model_width),
  model_height(model_height),
  x_sub(x_sub), y_sub(y_sub), theta_sub(theta_sub),
  z_max(feature_images.size()),
  feature_images(feature_images),
  transformation_space(boost::extents[x_sub][y_sub][theta_sub][z_max]),
  nb_thread(nb_thread)
#ifndef NDEBUG
  , display( &dpy )
#endif
{

  //Initiaise all threads ...
  for(int n = 0; n < nb_thread; n++)
    {
      image_space_vec.push_back(new boost::multi_array<std::vector<Transformation *>,2>
				(boost::extents[model_width + (x_max - x_min) + K][model_height + (y_max - y_min) + K]) );
      mutexes.push_back(new QMutex);
      mutexes_b.push_back(new QMutex);
      threads.push_back(new track_t(*image_space_vec[n], /*transf_it,*/ *mutexes[n], *mutexes_b[n]));
    }
  //... and launch them
  for(int n = 0; n < nb_thread; n++)
    {
      threads[n]->start();//QThread::TimeCriticalPriority);  ///think to join in destructor
    }

  float x_size = float(x_max - x_min)/ x_sub; //check this is not rounded
  float y_size = float(y_max - y_min)/ y_sub; 
  float theta_size = float(theta_max - theta_min)/ theta_sub; 

  //Fill the transformation space with its values
  for(idx_ts x = 0; x < x_sub; x++)
    for(idx_ts y = 0; y < y_sub; y++)
      for(idx_ts theta = 0; theta  < theta_sub; theta++)
	for(idx_ts z = 0; z < z_max; z++)
	  {
	    //add an error to see if we get more hit
	    float error = 0.;
	    std::vector<Interval> tmp_interv;
	    tmp_interv.push_back(Interval(x_min + x * x_size - error, x_min + (1+x)*x_size + error));
	    tmp_interv.push_back(Interval(y_min + y * y_size - error, y_min + (1+y)*y_size + error));
	    tmp_interv.push_back(Interval(theta_min + theta * theta_size - error, theta_min + (1+theta)*theta_size + error));

	    //the stack is treated slighly differently
	    tmp_interv.push_back(Interval(z));

	    Transformation tmp_transf;
	    tmp_transf.intervals = tmp_interv;

	    transformation_space[x][y][theta][z] = tmp_transf; 
	  }
  for( std::vector< std::vector<Feature> >::const_iterator image_it =
         feature_images.begin(); image_it != feature_images.end(); ++image_it )
    number_of_features.push_back((*image_it).size());
}
  

/*
  Pseudo-code 

  fill_image_space()
  {
  for each transformation
  for each feature of the image corresponding to the transformation{
  calculate locus of point according to the sub-transformed space;      
  add transformation reference to this locus to the image_space array;
  }
  }

  this function isn't called by the constructor to later be able to serialize the
  tracker (Since it is long to calculate the data structure)
*/

void hough_tracking::fill_image_space(void){

  int switch_p = 0;//for parallelisation

  for(idx_ts theta = 0; theta  < theta_sub; theta++) {

    std::vector< std::pair< float, float > > rotations;

    // Pre-compute rotations to save time.
    Interval *angles = &transformation_space[0][0][theta][0].intervals[2];
    for ( float t = angles->lower_bound; t<= angles->upper_bound; t++ ) {
      float radian = t * M_PI / 180.0;
      rotations.push_back( std::make_pair( cos( radian ), sin( radian ) ) );
    };
    for(idx_ts x = 0; x < x_sub; x++)
      for(idx_ts y = 0; y < y_sub; y++)
        for(idx_ts z = 0; z < z_max; z++) {

	  switch_p++; 

          Transformation *tmp_transf = &transformation_space[x][y][theta][z];
          // for(unsigned int n=0; n < transformation_space.num_elements(); n++) {
	  // Transformation tmp_transf =  transformation_space.data()[n];
	  //intervals[3] correspond to the number of the image.
	
	  for( std::vector<Feature>::const_iterator feature_it = (feature_images.begin() + tmp_transf->intervals[3].value)->begin();
	       feature_it != (feature_images.begin() + tmp_transf->intervals[3].value)->end();
	       feature_it++)
	    {
	      //calculate the locus for this feature for this transformation
	      //lets start with rotation we assume !!!that
	      //leap of one degre won't make us lose any pixel
	      //we use a set of feature position to avoid repetition

	      std::set<Feature, ltfeature> rotated_locus;

	      //!!round value to find the integer box
	      std::vector< std::pair< float, float > >::const_iterator rotation_index =
		rotations.begin();
	      for(float t = tmp_transf->intervals[2].lower_bound; t <= tmp_transf->intervals[2].upper_bound; t++)
		{
		  Feature tmp_feature(2);
		  tmp_feature(0) =
		    (int)( (*feature_it)(0) * rotation_index->first -
			   (*feature_it)(1) * rotation_index->second );
		  tmp_feature(1) =
		    (int)( (*feature_it)(0) * rotation_index->second +
			   (*feature_it)(1) * rotation_index->first );
		  rotated_locus.insert( tmp_feature );
		  rotation_index++;
		}
	    
	      std::set<Feature, ltfeature> x_trans_rot_locus;
	      for(int x = int(tmp_transf->intervals[0].lower_bound); x <= tmp_transf->intervals[0].upper_bound; x++)
		{
		  for( std::set<Feature, ltfeature>::const_iterator feat_it = rotated_locus.begin(); feat_it != rotated_locus.end(); feat_it++)
		    {
		      Feature tmp_feature(2);
		      tmp_feature(0) = (*feat_it)(0) + x;
		      tmp_feature(1) = (*feat_it)(1);
		      x_trans_rot_locus.insert(tmp_feature);
		    }
		}

	      std::set<Feature, ltfeature> xy_trans_rot_locus;
	      for(int y = int(tmp_transf->intervals[1].lower_bound); y <= tmp_transf->intervals[1].upper_bound; y++)
		{
		  for( std::set<Feature, ltfeature>::const_iterator feat_it =
			 x_trans_rot_locus.begin();
		       feat_it != x_trans_rot_locus.end(); feat_it++)
		    {
		      Feature tmp_feature(2);
		      tmp_feature(0) = (*feat_it)(0);
		      tmp_feature(1) = (*feat_it)(1) + y;
		      xy_trans_rot_locus.insert(tmp_feature);
		    }
		}

	      //for parallelisation we fill a stencil for each image space
	      //fill the image space with the pointer to the transformation
	      for( std::set<Feature, ltfeature>::const_iterator feat_it = xy_trans_rot_locus.begin();
		   feat_it != xy_trans_rot_locus.end();
		   feat_it++)
		{
		  //transform back to top left coordinate system
		  int x = (*feat_it)(0) + image_space_vec[0]->shape()[0]/2;
		  int y = image_space_vec[0]->shape()[1]/2 - (*feat_it)(1);
	      
		  //we don't want to add twice the same transform if
		  //the same sub-transformation bring two points
		  //to the same place	      
		  if(((*image_space_vec[switch_p % nb_thread])[x][y].empty()) || ((*image_space_vec[switch_p % nb_thread])[x][y].back() != tmp_transf))
		    {
		      (*image_space_vec[switch_p % nb_thread])[x][y].push_back(tmp_transf);
		    }	    
	      
		}
	    }
#ifndef NDEBUG	
	  if(!(switch_p%37)) //prime number to avoid get trapped on a pattern corresponding to the same model images.
	    {		
	      //display image space for this transformation
	      multi_array<bool,2> transf_im(extents[image_space_vec[0]->shape()[1]][image_space_vec[0]->shape()[0]]);
	  
	      for(idx_is x = 0; (unsigned int)(x) < image_space_vec[0]->shape()[0] ; x++)
		for(idx_is y = 0;(unsigned int) (y) < image_space_vec[0]->shape()[1] ; y++)
		  if((!(*image_space_vec[switch_p % nb_thread])[x][y].empty()) && ((*image_space_vec[switch_p % nb_thread])[x][y].back() == tmp_transf ))
		    {
		      transf_im[y][x] = true; 
		    }
	      display << multi_array_to_image( tobw( transf_im ) );
	    }
      
#endif
	}
  }
}


//For each feature, add a vote to the transformations that 
//bring a feature to this position. Then find a/the transformation with the
//maximum of vote.
//Note that features should be expressed relatively to the previous tracked location.

Transformation hough_tracking::track(std::vector<Feature> &features)
{
#ifndef NDEBUG
  QTime time;
  time.start();
#endif

  //reinitialize the number of vote
  for(unsigned int n = 0; n < transformation_space.num_elements(); n++)
    transformation_space.data()[n].votes = 0;

  //here use nb_thread threads to do the parallelization

  for (int i = 0; i < nb_thread; i++)
    {  
      threads[i]->set_features(features);
    }

#ifndef NDEBUG
  std::cerr<<"Nb of features in threads: "<< threads[0]->features.size()<<std::endl;
#endif

  //to make sure that the features are set properly
  for (int i = 0; i < nb_thread; i++)
    {  
      mutexes_b[i]->unlock();
    }

  //to synchronise so all the threads have finished their job before looking for the maximum
  for (int i = 0; i < nb_thread; i++)
    {  
      mutexes[i]->lock();
    }

  Transformation max_element = *(std::max_element(transformation_space.data(),
						  transformation_space.data()+ transformation_space.num_elements(),
						  lttransf()));

#ifndef NDEBUG
  std::cerr<<"Total time for tracking: "<< time.restart()<<std::endl;
#endif
  return max_element;
}
