#include "hough_tracking_data.hh"
#include "hough_tracking_plugin.hh"
#include "tools.hh"

using namespace boost;
using namespace boost::numeric;
using namespace mimas;
using namespace std;

#define VISIONTYPE "hough-tracking"
#define SLOWFEATURES

template<typename ImageIteratorSource>
std::vector<Feature> fast_features(ImageIteratorSource ul, ImageIteratorSource lr, unsigned int t)
{
  std::vector<Feature> features;
  ImageIteratorSource is(ul);

  unsigned int w = lr.x -ul.x , h = lr.y - ul.y;
  
  // we don't take into account the border cases
  is.y++; 
  for(unsigned int y = 1; y< h - 1; y++, is.y++)
    {
      //create an iterator for each line
      ImageIteratorSource ixs(is);

      // we don't take into account the border cases
      ixs.x++;
      for (unsigned int x = 1; x < w - 1; x++, ixs.x++)
	{		
	  const ImageIteratorSource ref(ixs);

	  //!!!Here I don't understand how it is working for
	  //color images as I haven't found any operator-
	  //for two rgba pixels. However it is working.
	  int temp1 = ref[dist2D(0,-1)] - ref[dist2D(0,1)];
	  int temp2 = ref[dist2D(-1,0)] - ref[dist2D(1,0)];
	  	  
	  if(temp1 < 0)
	    temp1 = -temp1;
	  if(temp2 < 0)
	    temp2 = -temp2;

	  if(unsigned(temp1 + temp2) > t)
	    {
	      Feature e(2);
	      e[0] = x;
	      e[1] = y;
	      features.push_back(e);
	    }
	}      
    }  
  return features;
}

hough_tracking_plugin::hough_tracking_plugin
  ( const calibration_ptr &_calib,
    const mimas::xml_reference< XalanElement > &configuration,
    bool _verbose ):
    recognition_method_4dof( _calib, Tracking, configuration, _verbose )
#ifndef NDEBUG
    , display( &dpy )
#endif
{
  //for each object type we are tracking a tracker is created.
  //((still) don't take into account multiple instances
  //of a same object type)
  xml_node_reference_list object_types = configuration.selectNodes
    ("database/object-type/tracking/"VISIONTYPE );

  for(int n = 0; n < object_types.getLength(); n++)
    {
      xml_reference< XalanElement > object_type =
        dynamic_pointer_cast< XalanElement >( object_types.item(n) );

      Tracker currentTracker;

      xml_node_reference_list file_names = object_type.selectNodes
        ("../../shared/stack/image");

      currentTracker.z_resolution =
        atof(object_type.selectNode("../../shared/stack/z-resolution").
             getNodeText().c_str() ) * MICROMETER;

      std:: string class_name =
        object_type.selectNode("../../@name").getNodeText();

      int sampling_rate = atoi( object_type.selectNode( "sampling_rate").
                                getNodeText().c_str() );

      xml_reference<XalanElement> model_param
        ( dynamic_pointer_cast< XalanElement >
          ( object_type.selectNode( "model" ) ) );

      int model_width = atoi(model_param.selectNode("width").getNodeText().c_str());
      int model_height = atoi(model_param.selectNode("height").getNodeText().c_str());
      int center_x = atoi(model_param.selectNode("center_x").getNodeText().c_str());
      int center_y = atoi(model_param.selectNode("center_y").getNodeText().c_str());
      double theta = atof(model_param.selectNode("angle").getNodeText().c_str());

      int min_gradient = atoi(model_param.selectNode("min_gradient").getNodeText().c_str());
      currentTracker.min_gradient = min_gradient;

      currentTracker.nb_fail = atoi(object_type.selectNode( "verification/nb_image").getNodeText().c_str());
      currentTracker.thresh = atof(object_type.selectNode( "verification/threshold").getNodeText().c_str());
      currentTracker.sampling_rate = sampling_rate;
      
      xml_reference<XalanElement> tracking_param =
        dynamic_pointer_cast< XalanElement >
        ( object_type.selectNode( "parameters" ) );


      int x_min = atoi(tracking_param.selectNode("x_min").getNodeText().c_str());
      int x_max = atoi(tracking_param.selectNode("x_max").getNodeText().c_str());
      int x_sub = atoi(tracking_param.selectNode("x_sub").getNodeText().c_str());
  
      int y_min = atoi(tracking_param.selectNode("y_min").getNodeText().c_str());
      int y_max = atoi(tracking_param.selectNode("y_max").getNodeText().c_str());
      int y_sub = atoi(tracking_param.selectNode("y_sub").getNodeText().c_str());
  
      float theta_min = (float)atof(tracking_param.selectNode("theta_min").getNodeText().c_str());
      float theta_max = (float)atof(tracking_param.selectNode("theta_max").getNodeText().c_str());
      int theta_sub = atoi(tracking_param.selectNode("theta_sub").getNodeText().c_str()); 

      std::vector< std::vector< Feature> > model_stack =
        create_model( sampling_rate,
                      model_width, model_height, center_x, center_y,
                      theta, min_gradient, file_names );

      currentTracker.tracker =
        hough_tracking_ptr
        ( new hough_tracking( model_stack, model_width, model_height, 40,
                              x_min,x_max,x_sub,
                              y_min,y_max,y_sub,
                              theta_min,theta_max,theta_sub,2) );

      custom_offset( class_name,
                     mi_pose( -theta * M_PI / 180.0, 0, 0,
                              center_x * sampling_rate *
                              calib->get_pixel_size(),
                              center_y * sampling_rate *
                              calib->get_pixel_size(),
                              0.0 ).inv() );
    
      currentTracker.tracker->fill_image_space();
      trackers[class_name] = currentTracker;
    }
}

std::vector< std::vector< Feature> >
hough_tracking_plugin::create_model( int sampling_rate,
                                     int model_width, int model_height,
                                     int center_x, int center_y,
                                     double theta, int min_gradient,
                                     xml_node_reference_list &file_names ){
  std::vector< std::vector< Feature> > model_stack;
  for( int i=0; i<file_names.getLength(); i++ )
    {
      image<unsigned char > image
        ( down_sample( load_image< unsigned char >
                       ( file_names.item( i ).getNodeText().c_str() ),
                       sampling_rate, 0 ) );
      
#ifdef SLOWFEATURES
      std::vector< Feature> points =
        featurePoints( edgeDetect( image, min_gradient ) );
#else
      std::vector< Feature> points = fast_features(image.ul(), image.lr(), min_gradient);
#endif
	
      std::vector< Feature> modelPoints;

      //express feature points in the rotated/translated object centered frame
      //If theta is the angle you need to rotate from the desired position of
      // the model to the actual position 
      // we need to rotate of -theta to put the model to the original
      // (desired) position.
      mi_pose transform( mi_pose::rotate_2d( theta * M_PI / 180.0 ) *
                         mi_pose::scale_2d( 1.0, -1.0 ) *
                         mi_pose::translate_2d( -center_x, -center_y ) );

      for( std::vector< Feature>::iterator feature_it = points.begin();
           feature_it != points.end(); feature_it++) {
        
        Feature mapped( transform.map( *feature_it ) );

        //eliminate features that are not in the object boundary
        if( mapped(0) >= -model_width/2  && mapped(0) < model_width/2 &&
            mapped(1) > -model_height/2 && mapped(1) <= model_height/2 ) {
          modelPoints.push_back(mapped);
        }
      }

#ifndef NDEBUG
	//display image space for this transformation
	multi_array<bool,2> transf_im(extents[500][500]);
	std::cerr<<"Number of points: "<<modelPoints.size()<<std::endl;	
	for(std::vector<Feature>::iterator it = modelPoints.begin(); it != modelPoints.end() ; it++)
	      {
		transf_im[(*it)(1)+250][(*it)(0)+250] = true; 
	      }
	display << multi_array_to_image( tobw( transf_im ) );
#endif

      model_stack.push_back(modelPoints);
    }
  return model_stack;
}
 

///This method is just interfacing the framework with the tracker.
vector< match > hough_tracking_plugin::recognise
( const image< unsigned char > &img, double timestamp,
  mi_scene_ptr scene ) const
{
  if ( verbose ) cout << "   "VISIONTYPE":" << endl;

  vector< match > retVal;

  //!!!carefull for the moment we suppose we haven't tested multiple objects

  //----load from the framework previous position of this object.----//
  mi_object_instance_ptr object; 
  for (int number = 0; number < (signed)scene->get_objects().size();
       number++) {

    object = scene->get_objects()[number];

    if ( trackers.find( object->get_type() ) != trackers.end() ) {

      if ( verbose )
        cout << "      object \"" << object->get_type() << "\":" << endl;

      const Tracker *current_tracker =
        &trackers.find( object->get_type() )->second;

      float z_resolution = current_tracker->z_resolution;
      int min_gradient = current_tracker->min_gradient;
      hough_tracking_ptr tracker = current_tracker->tracker;
  
      dof4 dof = convert_pose_to_4dof( object->get_type(), object->get_pose(),
                                       z_resolution );
      hough_tracking_data_ptr data;
      if ( object->data( Tracking ) )
        data = static_pointer_cast< hough_tracking_data >
          ( object->data( Tracking ) );

      //(x_position,y_position) is the pixel position of the
      //center of the image model in the image frame.
      //theta position is the angle given in degree in direct sense
      //of object with the image model both centers being matched 
      //z is the position of the image in the stack
      //0 being the 1st image and so on.

      // New code (requires different offset in XML file!)
      int x_position = (int)( dof.getX() / current_tracker->sampling_rate );
      int y_position = (int)( dof.getY() / current_tracker->sampling_rate );


      float theta_position = -dof.getAngle() / M_PI * 180.0;
      int z_position = dof.getZ();

#ifndef NDEBUG
      image<unsigned char> test_image;
      test_image = img;  
      test_image.setPixel(x_position * current_tracker->sampling_rate,
                          y_position * current_tracker->sampling_rate, 128);
      display << image<rgba<unsigned char > >(test_image)<<mimas::delay(500);
#endif


      //---------------get the features of the image and track the object ------------------//
      std::vector< Feature> im_features;
      if ( current_tracker->sampling_rate == 1 )
#ifdef SLOWFEATURES
        im_features = featurePoints( edgeDetect( img, min_gradient ) );
#else
        im_features = fast_features(img.ul(), img.lr(), min_gradient);
#endif
      else {
        image< unsigned char > sampled
          ( down_sample( img, current_tracker->sampling_rate, 0 ) );
#ifdef SLOWFEATURES
        im_features = featurePoints( edgeDetect( sampled, min_gradient ) );
#else
        im_features = fast_features(sampled.ul(), sampled.lr(), min_gradient);
#endif
      };
      
      std::vector< Feature> centered_features;

      //express image feature points in the rotated/translated object
      // centered frame	
      //center on the object
      // rotate of -theta    
      mi_pose transform( mi_pose::rotate_2d( -theta_position * M_PI / 180.0 ) *
                         mi_pose::scale_2d(  1.0 / calib->get_scale(),
                                            -1.0 / calib->get_scale() ) *
                         mi_pose::translate_2d( -x_position, -y_position ) );

      int
        model_width = tracker->get_model_width(),
        model_height = tracker->get_model_height();

      for( std::vector< Feature>::const_iterator feature_it =
             im_features.begin(); feature_it != im_features.end();
           feature_it++) {
        Feature mapped( transform.map( *feature_it ) );

        //eliminate features that are not in the object boundary 
        if( mapped(0) >= -model_width/2  && mapped(0) < model_width/2 &&
            mapped(1) > -model_height/2 && mapped(1) <= model_height/2 ) {
          centered_features.push_back(mapped);
        }
      }
      
      Transformation result = tracker->track(centered_features);
      
      float percentage = float(result.votes) / tracker->number_of_features[result.intervals[3].value];
#ifndef NDEBUG
      cerr<<"Indicator of success is "<<percentage<<endl;
#endif

      bool success;
      if( percentage > current_tracker->thresh ) {
        success = true;
        data = hough_tracking_data_ptr( new hough_tracking_data( 0 ) );
      } else {
        int old_nb_thresh_failed;
        if ( data )
          old_nb_thresh_failed = data->get_nb_thresh_failed();
        else
          old_nb_thresh_failed = 0;
        data = hough_tracking_data_ptr
          ( new hough_tracking_data( old_nb_thresh_failed + 1 ) );
        success = data->get_nb_thresh_failed() <= current_tracker->nb_fail;
      };

      if ( verbose ) {
        assert( data );
        cout << "         Threshold reached             : "
             << ( percentage > current_tracker->thresh ? "yes" : "no" )
             << " (quota of feature-hits was " << percentage << ")" << endl
             << "         Number of consecutive failures: "
             << data->get_nb_thresh_failed() << " ("
             << ( success ? "ok" : "tracking failed" ) << ")" << endl;
      };

#ifndef NDEBUG
      if ( !success )
        cerr<<"tracking of this object has failed"<<endl;
#endif
      
      if ( success ) {
	//----------Fill the values for the framework--------------//
  
	theta_position += (int)((result.intervals[2].lower_bound + result.intervals[2].upper_bound)/2); 
	z_position = result.intervals[3].value;

	//The translation found is expressed relatively to the frame centered on the previous
	//position of the object. We have to express this on the image frame thus the following
	//rotation.
	Feature shift(2);
	Feature shift2(2);
	
	shift(0) = (int)( (result.intervals[0].lower_bound + result.intervals[0].upper_bound) / 2.0 );
	shift(1) = (int)( (result.intervals[1].lower_bound + result.intervals[1].upper_bound) / 2.0 );
  
	mi_pose rotate_back(theta_position * M_PI/180, 0,0);
	shift2 = rotate_back.map(shift);

	x_position += shift2(0);
	y_position -= shift2(1);

	dof4 res( x_position * current_tracker->sampling_rate,
             y_position * current_tracker->sampling_rate,
             z_position,
             -theta_position * M_PI / 180.0 );

	mi_pose pose_res( convert_4dof_to_pose( object->get_type(), res,
                                           z_resolution ) );

	retVal.push_back( match( object->get_type() , pose_res, percentage,
                            data ) );
      }
	
    }
  };
  
  return retVal;
}
