#include <mimas/gauss.h>
#include <mimas/image_funcs.h>
#include <mimas/xml_node_reference_list.h>
#include "maxima_extraction_plugin.hh"

using namespace boost;
using namespace boost::numeric;
using namespace mimas;
using namespace std;

#define VISIONTYPE "maxima-extraction"

maxima_extraction_plugin::maxima_extraction_plugin
   ( const calibration_ptr &_calib,
     const mimas::xml_reference< XalanElement >
     &configuration,
     bool _verbose ):
     recognition_method( _calib, Recognition, configuration, _verbose )
{
  // Read global parameters.
  xml_reference< XalanElement >
    global_config( dynamic_pointer_cast< XalanElement >
                   ( configuration.selectNode( "methods/"VISIONTYPE ) ) );
  threshold = atoi( global_config.selectNode( "threshold" ).
                    getNodeText().c_str() );
  sigma = atoi( global_config.selectNode( "sigma" ).
                getNodeText().c_str() );

  xml_node_reference_list objects
    ( configuration.selectNodes
      ( "database/object-type/recognition/"VISIONTYPE ) );

  for ( int i=0; i<objects.getLength(); i++ ) {
    xml_reference< XalanElement > item
      ( dynamic_pointer_cast< XalanElement >( objects.item( i ) ) );
    string className( item.selectNode( "../../@name" ).getNodeText() );
    models.insert( className );
  };
}

vector< match > maxima_extraction_plugin::recognise
  ( const image< unsigned char > &img, double timestamp,
    mi_scene_ptr scene ) const
{
  if ( verbose ) cout << "   "VISIONTYPE":" << endl;

  vector< match > retVal;
  std::string activeModel;

  for ( map< string, int >::const_iterator i = scene->get_jobs().begin();
        i != scene->get_jobs().end(); i++ )
    if ( models.find( i->first ) != models.end() ) {
      MMERROR( activeModel.empty(), mimasexception, ,
               "maxima-extraction cannot distinguish between different types "
               "of objects (here: \"" << activeModel << "\" and \"" << i->first
               << "\"." );
      activeModel = i->first;
    };

  if ( !activeModel.empty() ) {
    
    if ( verbose ) cout << "      object \"" << activeModel << "\":" << endl;

    image< unsigned char > blurredImage
      ( gaussBlur< float >( img, sigma,
                               (float)( 8.0 / 256.0 ) ) );

    int numMaxima = 0;

    for ( int y=1; y<blurredImage.getHeight()-1; y++ )
      for ( int x=1; x<blurredImage.getWidth()-1; x++ ) 
        if ( blurredImage.pixel( x, y ) >= threshold ) {
          for ( int j=-1; j<=1; j++ )
            for ( int i=-1; i<=1; i++ ) {
              if ( j < 0 || ( j == 0 && i < 0 ) ) {
                if ( blurredImage.pixel( x + i, y + j ) >=
                     blurredImage.pixel( x, y ) )
                  goto exitLoop;
              } else
                if ( j != 0 || i != 0 )
                  if ( blurredImage.pixel( x + i, y + j ) >
                       blurredImage.pixel( x, y ) )
                    goto exitLoop;
            };

          // Add match emphasizing the ones in the middle of the picture.
          numMaxima++;
          retVal.push_back
            ( match( activeModel,
                     convert_pose_to_external( activeModel,
                                               pose3d( mi_pose( .0, x, y ) ) ),
                     1.0
                     - fabs( (double)x / blurredImage.getWidth() - 0.5 )
                     - fabs( (double)y / blurredImage.getHeight() - 0.5 ) ) );
        exitLoop:;
        };
    if ( verbose ) cout << "         number of maxima: " << numMaxima
                        << " (luminosity of filtered image up to "
                        << (int)max_val( blurredImage ) << ")" << endl;
  };

  return retVal;
}

mi_pose maxima_extraction_plugin::pose3d( const mi_pose &pose ) const
{
  assert( pose.getRotation().size1() == 2 );
  // It's better to check twice.
  assert( pose.getRotation().size1() == 2 );
  assert( pose.getTranslation().size() == 2 );

  ublas::matrix< double > rotation3d( 3, 3 );
  rotation3d( 0, 0 ) = pose.getRotation()( 0, 0 );
  rotation3d( 0, 1 ) = pose.getRotation()( 0, 1 );
  rotation3d( 1, 0 ) = pose.getRotation()( 1, 0 );
  rotation3d( 1, 1 ) = pose.getRotation()( 1, 1 );
  rotation3d( 2, 2 ) = 1.0;

  ublas::vector< double > translation3d( 3 );
  translation3d( 0 ) =
    ( pose.getTranslation()( 0 ) - calib->get_middle_x() ) *
    calib->get_pixel_size();
  translation3d( 1 ) =
    ( pose.getTranslation()( 1 ) - calib->get_middle_y() ) *
    calib->get_pixel_size();
  translation3d( 2 ) = 0.0;

  return mi_pose( rotation3d, translation3d );
}
