#ifndef __RECOGNITION_METHOD_H
#define __RECOGNITION_METHOD_H

#include <boost/smart_ptr.hpp>
#include <map>
#include <mimas/image.h>
#include <mimas/object.h>
#include <mimas/xml_reference.h>
#include "calibration.hh"
#include "match.hh"
#include "mi_scene.hh"
#include "vision_type.hh"

/** Abstract class for a vision-algorithm.
    All "plugin"-classes for recognition/tracking are deriving from this
    class.
    @author Jan Wedekind (j.wedekind@shu.ac.uk) */
class recognition_method: public mimas::object
{
 public:
  /** Constructor.
      The constructor stores a pointer to the calibration-data and it reads
      the customisation-transformations for the coordinate-systems of the
      micro-objects.
      @param _calib Calibration data.
      @param _type Determines, wether this is a tracking- or
      recognition-method.
      @param configuration Root-element of xml-document for configuration.
      @param _verbose Set to \c true, if you want to get verbose information
      from the vision-algorithms. */
  recognition_method( const calibration_ptr &_calib,
                      vision_type _type,
                      const mimas::xml_reference< XalanElement >
                      &configuration, bool _verbose );
  /** Virtual recognition method.
      This method has to be implemented by the deriving class.
      @param image Image to recognise objects in.
      @param timestamp Timestamp of image.
      @param scene Scene-object containing vision-jobs and position of tracked objects.
      @return List of hypotheses.
      @see match */
  virtual std::vector< match > recognise
    ( const mimas::image< unsigned char > &image,
      double timestamp,
      mi_scene_ptr scene ) const = 0;
  /** Convert internal pose of recognition method to external pose.
      @param type Type of micro-object.
      @param pose Internal pose of vision-algorithm. */
  mi_pose convert_pose_to_external( const std::string &type,
                                    const mi_pose &pose ) const
    throw (mimas::mimasexception);
  /** Convert external pose to internal pose of recognition method.
      @param type Type of micro-object.
      @param pose External pose of framework. */
  mi_pose convert_pose_to_internal( const std::string &type,
                                    const mi_pose &pose ) const
    throw (mimas::mimasexception);
  /// Get type of vision-method.
  vision_type get_type(void) const { return type; }
 protected:
  /// Prepend transform for internal customisation.
  void custom_offset( const std::string &type,
                      const mi_pose &pose );
  /// Calibration data.
  calibration_ptr calib;
  /// Customisation for pose of micro-objects.
  std::map< std::string, mi_pose > offset;
  /// Type of vision-method.
  vision_type type;
  /// Verbosity-switch.
  bool verbose;
};

/// Pointer to vision-algorithm object.
typedef boost::shared_ptr< recognition_method > recognition_method_ptr;

#endif
