#include "micron_vision_config.h"
#include <GL/glut.h>
#include <mimas/filter.h>
#include <mimas/image_filesinput.h>
#include <mimas/image_funcs.h>
#ifdef HAVE_LIBDC1394
#include <mimas/image_dc1394input.h>
#endif
#include <mimas/image_gstreamerinput.h>
#include <mimas/image_v4linput.h>
#include "tools.hh"
#include "units.h"

using namespace boost;
using namespace boost::numeric;
using namespace mimas;
using namespace std;

image< short int > gradX( const image< unsigned char > &m )
{
  image< short int > retVal;
  filter< short int >::gradX
    ( image< short int >( m ), retVal );
  return retVal;
}

image< short int > gradY( const image< unsigned char > &m )
{
  image< short int > retVal;
  filter< short int >::gradY
    ( image< short int >( m ), retVal );
  return retVal;
}

void dxvec( short int gx, short int gy, short int w,
            signed char &dx, short int &dy )
{
  short int
    agx = gx < 0 ? -gx : gx,
    agy = gy < 0 ? -gy : gy;
  if ( agy < agx ) {
    if ( ( agy << 1 ) < agx )
      dy = 0;
    else
      dy = gy < 0 ? -w : w;
    dx = gx < 0 ? -1 : 1;
  } else {
    if ( ( agx << 1 ) < agy )
      dx = 0;
    else
      dx = gx < 0 ? -1 : 1;
    dy = gy < 0 ? -w : w;
  };
}

image< bool > edgeDetect( const image< unsigned char > &m,
                             short int threshold )
{
  int thresholdSqr = (int)threshold * threshold;

  image< short int >
    gradientX( gradX( m ) ),
    gradientY( gradY( m ) );

  image< int > 
    gradientSqr( fastSqr( gradientX ) + fastSqr( gradientY ) );

  image< bool > retVal; retVal.init( m.getWidth(), m.getHeight() );

  short int w = (short)m.getWidth();
  const int *pgSqr = gradientSqr.rawData().data() + w + 1;
  const short int
    *pgX = gradientX.rawData().data() + w + 1,
    *pgY = gradientY.rawData().data() + w + 1;
  bool *pr = retVal.rawData().data() + w + 1;

  for ( int y=1; y<m.getHeight()-1; y++ ) {
    for ( int x=1; x<m.getWidth()-1; x++ ) {

      signed char dx;
      short int dy;

      dxvec( *pgX, *pgY, m.getWidth(), dx, dy );

      int g = *pgSqr;
      if ( g >= thresholdSqr && 
           g >= pgSqr[ - ( dy + dx ) ] &&
           g >= pgSqr[ + ( dy + dx ) ] )
        *pr = true;

      pgSqr++;
      pgX++;
      pgY++;
      pr++;
    };
    pgSqr += 2;
    pgX += 2;
    pgY += 2;
    pr += 2;
  };

  return retVal;
}

image< unsigned char > down_sample( const image< unsigned char > &x,
                                       int rate, int fringe )
{
  int
    h = ( x.getHeight() + rate - 1 - 2 * fringe ) / rate,
    w = ( x.getWidth()  + rate - 1 - 2 * fringe ) / rate;
  multi_array< unsigned char, 2 > retVal( extents[ h ][ w ] );
  typedef multi_array< double, 2 > array_type;
  typedef array_type::index_range range;
  array_type::index_gen indices;
  retVal = x.rawData()[ indices[ range( fringe, rate * h + fringe, rate ) ]
                               [ range( fringe, rate * w + fringe, rate ) ] ];
  return multi_array_to_image( retVal );
}

image< bool > gradientLevels
   ( const image< unsigned char > &image,
     short int minGradient )
{
  return bilevel< bool >( fastSqr( gradX( image ) ) +
                          fastSqr( gradY( image ) ),
                          (int)minGradient * minGradient,
                          false, true );
}

ublas::vector< int > center( const vector< ublas::vector< int > > &points )
{
  ublas::vector< int > retVal( 2 );

  if ( !points.empty() ) {

    ublas::vector< int > sum( 2 );

    for ( vector< ublas::vector< int > >::const_iterator i = points.begin();
          i != points.end(); i++ )
      sum += *i;
    
    retVal = sum / points.size();
  };

  return retVal;
}

vector< ublas::vector< int > > featurePoints
   ( const image< bool > &binary )
{
  vector< ublas::vector< int > > retVal;
  const bool *p = binary.rawData().data();
  for ( int y=0; y<(signed)binary.getHeight(); y++ )
    for ( int x=0; x<(signed)binary.getWidth(); x++ )
      if ( *p++ ) {
        ublas::vector< int > e( 2 );
        e[0] = x;
        e[1] = y;
        retVal.push_back( e );
      };
  return retVal;
}

vector< ublas::vector< int > > subtractOffset
  ( const vector< ublas::vector< int > > &points,
    const ublas::vector< int > &offset )
{
  vector< ublas::vector< int > > retVal;
  for ( vector< ublas::vector< int > >::const_iterator i = points.begin();
        i != points.end(); i++ ) {
    retVal.push_back( *i - offset );
  };
  return retVal;  
}

vector< ublas::vector< int > > scalePoints
  ( const vector< ublas::vector< int > > &points,
    double scale )
{
  vector< ublas::vector< int > > retVal;
  if ( scale == 1.0 )
    retVal = points;
  else {
    for ( vector< ublas::vector< int > >::const_iterator i = points.begin();
          i != points.end(); i++ )
      retVal.push_back( *i * scale );
  };
  return retVal;  
}

shared_ptr< image_input< rgba< unsigned char > > > open_image_input
  ( int selectedInput, poptContext &optCon,
    int gstreamerSHMKey, int gstreamerSEMKey,
    const char *device, int channel,
    int width, int height, int brightness, int contrast,
    int shutter, int gain, int balance )
{
  shared_ptr< image_input< rgba< unsigned char > > > retVal;

  switch ( selectedInput ) {
  case 0: {
    // Get next stray command-line parameter.
    const char *inputFileName = poptGetArg( optCon );
    // Not a single input filename was specified :-(
    MMERROR( inputFileName != NULL, mimasexception, ,
             "No image-input was specified" );
    shared_ptr< image_filesinput< rgba< unsigned char > > >
      filesInput( new image_filesinput< rgba<  unsigned char > > );
    do {
      filesInput->addFileName( inputFileName );
      inputFileName = poptGetArg( optCon );
    } while ( inputFileName != NULL );
    retVal = filesInput;
    break; }
  case 1: {
    // No file-names on the command-line are allowed in gstreamer-mode.
    const char *strayParameter = poptGetArg( optCon );
    MMERROR( strayParameter == NULL, mimasexception, ,
             "Unexpected command-line argument \"" << strayParameter
             << "\"" );
    shared_ptr< image_gstreamerinput< rgba< unsigned char > > >
      gstreamerInput
      ( new image_gstreamerinput< rgba< unsigned char > >
        ( gstreamerSHMKey, gstreamerSEMKey ) );
    retVal = gstreamerInput;
    break; }
  case 2: {
    // No file-names on the command-line are allowed in video4linux-mode.
    const char *strayParameter = poptGetArg( optCon );
    MMERROR( strayParameter == NULL, mimasexception, ,
             "Unexpected command-line argument \"" << strayParameter
             << "\"" );
    MMERROR( brightness >= 0 && brightness <= 65535, mimasexception, ,
             "Brightness must be in [0..65535], but was " << brightness
             << "." );
    MMERROR( contrast >= 0 && contrast <= 65535, mimasexception, ,
             "Contrast must be in [0..65535], but was " << contrast
             << "." );
    shared_ptr< image_v4linput< rgba< unsigned char > > >
      v4lInput( new image_v4linput< rgba< unsigned char > >
                ( device == NULL ? "/dev/video" : device, channel,
                  width, height ) );
    v4lInput->setSensivity( brightness, // brightness
                            32768,      // hue
                            32768,      // colour
                            contrast ); // contrast
				
    retVal = v4lInput;
    break; };
#ifdef HAVE_LIBDC1394
  case 3: {
    // No file-names on the command-line are allowed in video1394-mode.
    const char *strayParameter = poptGetArg( optCon );
    MMERROR( strayParameter == NULL, mimasexception, ,
             "Unexpected command-line argument \"" << strayParameter
             << "\"" );   

    int mode = -1;
    switch ( width ) {
    case 320:
      MMERROR( height == 240 || height == -1, mimasexception, ,
               "Only resolution 640x480 and 320x240 are supported at the "
               "moment." );
      mode = MODE_320x240_YUV422;
      break;
    case -1:
    case 640:
      MMERROR( height == 480 || height == -1, mimasexception, ,
               "Only resolution 640x480 and 320x240 are supported at the "
               "moment." );
      mode = MODE_640x480_YUV422;
      break;
    default:
      MMERROR( false, mimasexception, ,
               "Only resolution 640x480 and 320x240 are supported at the "
               "moment." );
    };
    assert( mode != -1 );

    shared_ptr< image_dc1394input< rgba< unsigned char > > >
      dcInput( new image_dc1394input< rgba< unsigned char > >
               ( mode, FRAMERATE_15, channel, device ) );

    dc1394_feature_info shutterInfo = dcInput->get_feature( FEATURE_SHUTTER );
    dcInput->set_feature_value( FEATURE_SHUTTER, shutter );
    if ( shutter == -1 ) shutter = ( shutterInfo.min + shutterInfo.max ) / 2;
    MMERROR( shutterInfo.min <= shutter && shutterInfo.max >= shutter,
             mimasexception, ,
             "Value for shutter must be between " << shutterInfo.min << " and "
             << shutterInfo.max );

    dc1394_feature_info gainInfo = dcInput->get_feature( FEATURE_GAIN );
    if ( gain == -1 ) gain = ( gainInfo.min + gainInfo.max ) / 2;
    MMERROR( gainInfo.min <= gain && gainInfo.max >= gain,
             mimasexception, ,
             "Value for gain must be between " << gainInfo.min << " and "
             << gainInfo.max );
    dcInput->set_feature_value( FEATURE_GAIN, gain );

    dc1394_feature_info balanceInfo = dcInput->
      get_feature( FEATURE_WHITE_BALANCE );
    if ( balance == -1 ) balance = ( balanceInfo.min + balanceInfo.max ) / 2;
    MMERROR( balanceInfo.min <= balance && balanceInfo.max >= balance,
             mimasexception, ,
             "Value for white balance must be between "
             << balanceInfo.min << " and " << balanceInfo.max );
    dcInput->set_feature_value( FEATURE_WHITE_BALANCE, balance );

    retVal = dcInput;
    break; };
#endif
  default:
    assert(false);
  };
  return retVal;
}
