#ifndef __TOOLS_H
#define __TOOLS_H

#include <boost/shared_ptr.hpp>
#include <boost/multi_array.hpp>
#include <boost/numeric/ublas/vector.hpp>
#include <complex>
#include <functional>
#include <mimas/image.h>
#include <mimas/image_input.h>
#include <mimas/multi_array_op.h>
#include <mimas/rgba.h>
#include <popt.h>
#include <string>

// #define MMUSEOSMESA

#ifndef MMUSEOSMESA
#define GLX_GLXEXT_PROTOTYPES
#include <GL/glx.h>
#include <X11/Xlib.h>
#else
#include <GL/osmesa.h>
#endif

/** Convert boost-array to mimas-image.
    @param x The 2-dimensional array.
    @return Resulting image of conversion.
    @author Jan Wedekind (j.wedekind@shu.ac.uk) */
template< typename T >
mimas::image< T > multi_array_to_image
   ( const boost::multi_array< T, 2 > &x );

/** Convert mimas-image to boost-array.
    @param x The image.
    @return Resulting 2-d array of conversion.
    @author Jan Wedekind (j.wedekind@shu.ac.uk) */
template< typename T >
boost::multi_array< T, 2 > image_to_multi_array
   ( const mimas::image< T > &x );

/// Random-number generator.
template< typename T >
struct rand_function: public std::unary_function< T, int >
{
  /// Constructor.
  rand_function(void) { srand( time( 0 ) ); }
  /// Function.
  int operator()( const T &size ) const {
    return (int)( rand() / ( RAND_MAX + 1.0 ) * size );
  }
};

/// Wrapper for element-wise conversion of bilevel-images to b/w-images.
template< typename T >
struct _tobw: public std::unary_function< T,
                                          unsigned char > {
  /// Map boolean to {0,255}.
  unsigned char operator()( const T &x ) const
  { return x ? 255 : 0; }
};

inline mimas::image< unsigned char > tobw( const mimas::image< bool > &im )
{
  return mimas::image< unsigned char >
    ( mimas::multi_func< unsigned char >( im.rawData(),
                                          _tobw< bool >() ) );
}

/// X-gradient on images.
mimas::image< short int > gradX
  ( const mimas::image< unsigned char > &m );
/// Y-gradient on images.
mimas::image< short int > gradY
  ( const mimas::image< unsigned char > &m );
/// Load image from file.
template< typename T >
mimas::image< T > load_image( const std::string &fileName );
/** Detect edges.
    Similar to canny edge-detector but faster and less accurate.
    http://www.pages.drexel.edu/~weg22/can_tut.html
    @author Jan Wedekind (j.wedekind@shu.ac.uk) */
mimas::image< bool > edgeDetect
  ( const mimas::image< unsigned char > &m,
    short int threshold );
/// Downsampling and fringe-removal.
mimas::image< unsigned char > down_sample
  ( const mimas::image< unsigned char > &x, int rate, int fringe );
/// Compute thresholded gradient.
mimas::image< bool > gradientLevels
  ( const mimas::image< unsigned char > &image,
  short int minGradient );
///
boost::numeric::ublas::vector< int > center
  ( const std::vector< boost::numeric::ublas::vector< int > > &points );
/// Convert bilevel-image to list of points.
std::vector< boost::numeric::ublas::vector< int > > featurePoints
  ( const mimas::image< bool > &binary );
/** Move points.
    Subtract an offset from each point in the list.
    @param points List of points.
    @param offset Offset to be subtracted. */
std::vector< boost::numeric::ublas::vector< int > > subtractOffset
  ( const std::vector< boost::numeric::ublas::vector< int > > &points,
    const boost::numeric::ublas::vector< int > &offset );

/** Scale points.
    Scale coordinates of points with a factor.
    @param scale Scaling factor. */
std::vector< boost::numeric::ublas::vector< int > > scalePoints
  ( const std::vector< boost::numeric::ublas::vector< int > > &points,
    double scale );

/// Function for opening an arbitrary input source for colour-images.
boost::shared_ptr< mimas::image_input< mimas::rgba< unsigned char > > >
   open_image_input
   ( int selectedInput, poptContext &optCon,
     int gstreamerSHMKey, int gstreamerSEMKey,
     const char *device, int channel,
     int width, int height, int brightness, int contrast,
     int shutter, int gain, int balance );

#include "tools.tcc"

#endif
